# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Provide new db schema for verification component

Revision ID: 484cd9413e66
Revises: e654a0648db0
Create Date: 2016-11-04 17:04:24.614075

"""

import datetime as dt

from alembic import op
import sqlalchemy as sa

from rally.common.db import models
from rally.common.db import sa_types
from rally import exceptions

# revision identifiers, used by Alembic.
revision = "484cd9413e66"
down_revision = "e654a0648db0"
branch_labels = None
depends_on = None


TASK_STATUSES = ["aborted", "aborting", "cleaning up", "failed", "finished",
                 "init", "paused", "running", "setting up", "soft_aborting",
                 "verifying"]

_MAP_OLD_TO_NEW_TEST_STATUSES = {
    "OK": "success",
    "FAIL": "fail",
    "SKIP": "skip"
}


verification_helper = sa.Table(
    "verifications",
    sa.MetaData(),
    sa.Column("id", sa.Integer, primary_key=True, autoincrement=True),
    sa.Column("uuid", sa.String(36), nullable=False),
    sa.Column("deployment_uuid", sa.String(36), nullable=False),
    sa.Column("status", sa.Enum(*TASK_STATUSES, name="enum_tasks_status"),
              default="init", nullable=False),
    sa.Column("set_name", sa.String(20)),
    sa.Column("tests", sa.Integer),
    sa.Column("errors", sa.Integer),
    sa.Column("failures", sa.Integer),
    sa.Column("time", sa.Float),
    sa.Column("created_at", sa.DateTime),
    sa.Column("updated_at", sa.DateTime)
)


results_helper = sa.Table(
    "verification_results",
    sa.MetaData(),
    sa.Column("id", sa.Integer, primary_key=True, autoincrement=True),
    sa.Column("verification_uuid", sa.String(36), nullable=False),
    sa.Column("data", sa_types.MutableJSONEncodedDict, nullable=False,
              default={}),
    sa.Column("created_at", sa.DateTime),
    sa.Column("updated_at", sa.DateTime)
)


def upgrade():
    connection = op.get_bind()

    # create new table to store all verifiers
    verifiers_table = op.create_table(
        "verifiers",
        sa.Column("id", sa.Integer, primary_key=True, autoincrement=True),
        sa.Column("uuid", sa.String(36), nullable=False),

        sa.Column("name", sa.String(255), unique=True),
        sa.Column("description", sa.Text),

        sa.Column("type", sa.String(255), nullable=False),
        sa.Column("namespace", sa.String(255)),

        sa.Column("source", sa.String(255)),
        sa.Column("version", sa.String(255)),
        sa.Column("system_wide", sa.Boolean),

        sa.Column("status", sa.String(36), default="init", nullable=False),

        sa.Column("extra_settings", sa_types.MutableJSONEncodedDict,
                  nullable=True),

        sa.Column("created_at", sa.DateTime),
        sa.Column("updated_at", sa.DateTime)
    )

    op.create_index("verifier_uuid", "verifiers", ["uuid"], unique=True)

    verifications_table = op.create_table(
        "verifications_new",
        sa.Column("id", sa.Integer, primary_key=True, autoincrement=True),
        sa.Column("uuid", sa.String(36), nullable=False),

        sa.Column("verifier_uuid", sa.String(36), nullable=False),
        sa.Column("deployment_uuid", sa.String(36), nullable=False),

        sa.Column("run_args", sa_types.MutableJSONEncodedDict),

        sa.Column("status", sa.String(36), default="init", nullable=False),

        sa.Column("tests_count", sa.Integer, default=0),
        sa.Column("failures", sa.Integer, default=0),
        sa.Column("skipped", sa.Integer, default=0),
        sa.Column("success", sa.Integer, default=0),
        sa.Column("unexpected_success", sa.Integer, default=0),
        sa.Column("expected_failures", sa.Integer, default=0),
        sa.Column("tests_duration", sa.Float, default=0.0),

        sa.Column("tests", sa_types.MutableJSONEncodedDict, default={}),

        sa.Column("created_at", sa.DateTime),
        sa.Column("updated_at", sa.DateTime),

        sa.ForeignKeyConstraint(["verifier_uuid"], ["verifiers.uuid"]),
        sa.ForeignKeyConstraint(["deployment_uuid"], ["deployments.uuid"])
    )

    default_verifier = None
    for vresult in connection.execute(results_helper.select()):
        if default_verifier is None:
            vuuid = models.UUID()
            connection.execute(
                verifiers_table.insert(),
                [{
                    "uuid": vuuid,
                    "name": "DefaultTempestVerifier",
                    "description": "It is the default verifier to assign all "
                                   "migrated verification results for",
                    "type": "tempest",
                    "namespace": "openstack",
                    "source": "n/a",
                    "version": "n/a",
                    "system_wide": False,
                    "status": "init",
                    "created_at": dt.datetime.utcnow(),
                    "updated_at": dt.datetime.utcnow()
                }]
            )
            default_verifier = connection.execute(
                verifiers_table.select().where(
                    verifiers_table.c.uuid == vuuid)).first()

        data = vresult.data
        if "errors" in data:
            # it is a very old format...
            for test in data["test_cases"].keys():
                old_status = data["test_cases"][test]["status"]
                new_status = _MAP_OLD_TO_NEW_TEST_STATUSES.get(
                    old_status, old_status.lower())
                data["test_cases"][test]["status"] = new_status

                if "failure" in data["test_cases"][test]:
                    data["test_cases"][test]["traceback"] = data[
                        "test_cases"][test]["failure"]["log"]
                    data["test_cases"][test].pop("failure")

        verifications = connection.execute(
            verification_helper.select().where(
                verification_helper.c.uuid == vresult.verification_uuid))
        # for each verification result we have single verification object
        verification = verifications.first()

        connection.execute(
            verifications_table.insert(),
            [{"uuid": verification.uuid,
              "verifier_uuid": default_verifier.uuid,
              "deployment_uuid": verification.deployment_uuid,
              "run_args": {"pattern": "set=%s" % verification.set_name},
              "status": verification.status,
              "tests": data["test_cases"],
              "tests_count": data["tests"],
              "failures": data["failures"],
              "skipped": data["skipped"],
              "success": data["success"],
              "unexpected_success": data.get("unexpected_success", 0),
              "expected_failures": data.get("expected_failures", 0),
              "tests_duration": data["time"],
              "created_at": vresult.created_at,
              "updated_at": vresult.updated_at
              }])

    op.drop_table("verification_results")
    op.drop_table("verifications")
    op.rename_table("verifications_new", "verifications")

    op.create_index(
        "verification_uuid", "verifications", ["uuid"], unique=True)


def downgrade():
    raise exceptions.DowngradeNotSupported()



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from rally import exceptions
from tests.unit import test


class ExceptionTestCase(test.TestCase):

    @mock.patch("rally.exceptions._exception_map")
    def test_find_exception(self, mock__exception_map):
        mock_response = mock.Mock()
        exc_class = mock.Mock()
        mock__exception_map.get.return_value = exc_class
        mock_response.json.return_value = {
            "error": {"args": "args", "msg": "msg"}
        }

        exc_instance = exceptions.find_exception(mock_response)
        self.assertEqual(exc_instance, exc_class.return_value)

        mock__exception_map.reset_mock()
        exc_class.reset_mock()
        mock_response.reset_mock()
        mock_response.json.return_value = {
            "error": {"args": None, "msg": "msg"}
        }
        exc_instance = exceptions.find_exception(mock_response)
        self.assertEqual(exc_instance, exc_class.return_value)

    def test_make_exception(self):
        exc = exceptions.RallyException("exc")
        self.assertEqual(exc, exceptions.make_exception(exc))

        mock_exc = mock.Mock()
        self.assertIsInstance(exceptions.make_exception(mock_exc),
                              exceptions.RallyException)



# Copyright (C) 2014 Yahoo! Inc. All Rights Reserved.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import copy
import operator
import re

from rally.common import logging
from rally.common.plugin import plugin
from rally import exceptions
from rally.task import scenario


LOG = logging.getLogger(__name__)


def convert(**kwargs):
    """Decorator to define resource transformation(s) on scenario parameters.

    The ``kwargs`` passed as arguments are used to map a key in the
    scenario config to the resource type plugin used to perform a
    transformation on the value of the key. For instance:

        @types.convert(image={"type": "glance_image"})

    This would convert the ``image`` key in the scenario configuration
    to a Glance image by using the ``glance_image`` resource
    plugin. Currently ``type`` is the only recognized key, but others
    may be added in the future.
    """

    def wrapper(cls):
        for k, v in kwargs.items():
            if "type" not in v:
                LOG.warning(
                    "The configuration for preprocessing '%s' argument of"
                    " %s Scenario plugin is wrong. Check documentation "
                    "for more details." % (k, cls.get_name()))

        cls._meta_setdefault("preprocessors", {})
        cls._meta_get("preprocessors").update(kwargs)
        return cls
    return wrapper


def preprocess(name, context, args):
    """Run preprocessor on scenario arguments.

    :param name: Scenario plugin name
    :param context: dict with contexts data
    :param args: Scenario arguments for the workload

    :returns processed_args: dictionary object with additional client
                             and resource configuration

    """
    preprocessors = scenario.Scenario.get(name)._meta_get(
        "preprocessors", default={})

    processed_args = copy.deepcopy(args)

    cache = {}
    resource_types = {}
    for src, type_cfg in preprocessors.items():
        if type_cfg["type"] not in resource_types:
            resource_cls = ResourceType.get(type_cfg["type"])
            resource_types[type_cfg["type"]] = resource_cls(context, cache)

        preprocessor = resource_types[type_cfg["type"]]
        if src in processed_args:
            res = preprocessor.pre_process(
                resource_spec=processed_args[src], config=type_cfg)
            if res is not None:
                processed_args[src] = res
    return processed_args


@plugin.base()
class ResourceType(plugin.Plugin, metaclass=abc.ABCMeta):
    """A helper plugin for pre-processing input data of resources."""

    def __init__(self, context, cache=None):
        self._context = context
        self._global_cache = cache if cache is not None else {}
        self._global_cache.setdefault(self.get_name(), {})
        self._cache = self._global_cache[self.get_name()]

    @abc.abstractmethod
    def pre_process(self, resource_spec, config):
        """Pre-process resource.

        :param resource_spec: A specification of the resource from the task
        :param config: A configuration for preprocessing taken from the plugin
        """


def obj_from_name(resource_config, resources, typename):
    """Return the resource whose name matches the pattern.

    resource_config has to contain `name`, as it is used to lookup a resource.
    Value of the name will be treated as regexp.

    An `InvalidScenarioArgument` is thrown if the pattern does
    not match unambiguously.

    :param resource_config: resource to be transformed
    :param resources: iterable containing all resources
    :param typename: name which describes the type of resource

    :returns: resource object uniquely mapped to `name` or `regex`
    """
    if "name" in resource_config:
        # In a case of pattern string exactly matches resource name
        matching_exact = [resource for resource in resources
                          if resource.name == resource_config["name"]]
        if len(matching_exact) == 1:
            return matching_exact[0]
        elif len(matching_exact) > 1:
            raise exceptions.InvalidScenarioArgument(
                "{typename} with name '{pattern}' "
                "is ambiguous, possible matches "
                "by id: {ids}".format(typename=typename.title(),
                                      pattern=resource_config["name"],
                                      ids=", ".join(map(
                                                    operator.attrgetter("id"),
                                                    matching_exact))))
        # Else look up as regex
        patternstr = resource_config["name"]
    elif "regex" in resource_config:
        patternstr = resource_config["regex"]
    else:
        raise exceptions.InvalidScenarioArgument(
            "{typename} 'id', 'name', or 'regex' not found "
            "in '{resource_config}' ".format(typename=typename.title(),
                                             resource_config=resource_config))

    pattern = re.compile(patternstr)
    matching = [resource for resource in resources
                if re.search(pattern, resource.name or "")]
    if not matching:
        raise exceptions.InvalidScenarioArgument(
            "{typename} with pattern '{pattern}' not found".format(
                typename=typename.title(), pattern=pattern.pattern))
    elif len(matching) > 1:
        raise exceptions.InvalidScenarioArgument(
            "{typename} with name '{pattern}' is ambiguous, possible matches "
            "by id: {ids}".format(typename=typename.title(),
                                  pattern=pattern.pattern,
                                  ids=", ".join(map(operator.attrgetter("id"),
                                                    matching))))
    return matching[0]


def obj_from_id(resource_config, resources, typename):
    """Return the resource whose name matches the id.

    resource_config has to contain `id`, as it is used to lookup a resource.

    :param resource_config: resource to be transformed
    :param resources: iterable containing all resources
    :param typename: name which describes the type of resource

    :returns: resource object mapped to `id`
    """
    if "id" in resource_config:
        matching = [resource for resource in resources
                    if resource.id == resource_config["id"]]
        if len(matching) == 1:
            return matching[0]
        else:
            raise exceptions.InvalidScenarioArgument(
                "{typename} with id '{id}' not found".format(
                    typename=typename.title(), id=resource_config["id"]))
    else:
        raise exceptions.InvalidScenarioArgument(
            "{typename} 'id' not found in '{resource_config}'".format(
                typename=typename.title(), resource_config=resource_config))


def _id_from_name(resource_config, resources, typename, id_attr="id"):
    """Return the id of the resource whose name matches the pattern.

    resource_config has to contain `name`, as it is used to lookup an id.
    Value of the name will be treated as regexp.

    An `InvalidScenarioArgument` is thrown if the pattern does
    not match unambiguously.

    :param resource_config: resource to be transformed
    :param resources: iterable containing all resources
    :param typename: name which describes the type of resource
    :param id_attr: id or uuid should be returned

    :returns: resource id uniquely mapped to `name` or `regex`
    """
    try:
        return getattr(obj_from_name(resource_config, resources, typename),
                       id_attr)
    except AttributeError:
        raise exceptions.RallyException(
            "There is no attribute {attr} in the object {type}".format(
                attr=id_attr, type=typename))


def _name_from_id(resource_config, resources, typename):
    """Return the name of the resource which has the id.

    resource_config has to contain `id`, as it is used to lookup a name.

    :param resource_config: resource to be transformed
    :param resources: iterable containing all resources
    :param typename: name which describes the type of resource

    :returns: resource name mapped to `id`
    """
    return obj_from_id(resource_config, resources, typename).name



# Copyright (C) 2014 Yahoo! Inc. All Rights Reserved.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from rally import exceptions
from rally.plugins.common import types
from tests.unit import test


class PathOrUrlTestCase(test.TestCase):

    @mock.patch("os.path.isfile")
    @mock.patch("requests.head")
    def test_preprocess_file(self, mock_requests_head, mock_isfile):
        mock_isfile.return_value = True
        path = types.PathOrUrl({}, {}).pre_process("fake_path", {})
        self.assertEqual("fake_path", path)

    @mock.patch("os.path.isfile")
    @mock.patch("requests.head")
    def test_preprocess_bogus(self, mock_requests_head, mock_isfile):
        mock_isfile.return_value = False
        mock_requests_head.return_value = mock.Mock(status_code=500)
        self.assertRaises(exceptions.InvalidScenarioArgument,
                          types.PathOrUrl({}, {}).pre_process,
                          "fake_path", {})
        mock_requests_head.assert_called_once_with(
            "fake_path", verify=False, allow_redirects=True)

    @mock.patch("os.path.isfile")
    @mock.patch("requests.head")
    def test_preprocess_url(self, mock_requests_head, mock_isfile):
        mock_isfile.return_value = False
        mock_requests_head.return_value = mock.Mock(status_code=200)
        path = types.PathOrUrl({}, {}).pre_process("fake_url", {})
        self.assertEqual("fake_url", path)


class FileTypeTestCase(test.TestCase):

    @mock.patch("builtins.open",
                side_effect=mock.mock_open(read_data="file_context"),
                create=True)
    def test_preprocess_by_path(self, mock_open):
        resource_spec = "file.yaml"
        file_context = types.FileType({}, {}).pre_process(resource_spec, {})
        self.assertEqual("file_context", file_context)

    @mock.patch("builtins.open", side_effect=IOError, create=True)
    def test_preprocess_by_path_no_match(self, mock_open):
        resource_spec = "nonexistant.yaml"
        self.assertRaises(IOError,
                          types.FileType({}, {}).pre_process,
                          resource_spec=resource_spec,
                          config={})


class FileTypeDictTestCase(test.TestCase):

    @mock.patch("builtins.open",
                side_effect=mock.mock_open(read_data="file_context"),
                create=True)
    def test_preprocess_by_path(self, mock_open):
        resource_spec = ["file.yaml"]
        file_context = types.FileTypeDict({}, {}).pre_process(resource_spec,
                                                              {})
        self.assertEqual({"file.yaml": "file_context"}, file_context)

    @mock.patch("builtins.open", side_effect=IOError, create=True)
    def test_preprocess_by_path_no_match(self, mock_open):
        resource_spec = ["nonexistant.yaml"]
        self.assertRaises(IOError,
                          types.FileTypeDict({}, {}).pre_process,
                          resource_spec=resource_spec, config={})



# Copyright 2014: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import jinja2

from rally.ui import utils
from tests.unit import test


class ModuleTestCase(test.TestCase):

    def test_get_template(self):
        template = utils.get_template("base.html")
        self.assertIsInstance(template,
                              jinja2.environment.Template)
        self.assertEqual("base.html", template.name)
        self.assertIn("include_raw_file", template.globals)

    def test_get_template_raises(self):
        self.assertRaises(jinja2.exceptions.TemplateNotFound,
                          utils.get_template, "nonexistent")



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Change verification statuses

Revision ID: f33f4610dcda
Revises: a6f364988fc2
Create Date: 2017-01-23 13:56:30.999593

"""

from alembic import op
import sqlalchemy as sa

from rally import exceptions

# revision identifiers, used by Alembic.
revision = "f33f4610dcda"
down_revision = "a6f364988fc2"
branch_labels = None
depends_on = None


verifications_helper = sa.Table(
    "verifications",
    sa.MetaData(),
    sa.Column("id", sa.Integer, primary_key=True, autoincrement=True),
    sa.Column("failures", sa.Integer, default=0),
    sa.Column("unexpected_success", sa.Integer, default=0),
    sa.Column("status", sa.String(36), nullable=False)
)


def upgrade():
    connection = op.get_bind()
    for v in connection.execute(verifications_helper.select()):
        new_status = v.status
        if v.status == "finished" and (
                v.failures != 0 or v.unexpected_success != 0):
            new_status = "failed"
        elif v.status == "failed":
            new_status = "crashed"
        else:
            pass

        if new_status != v.status:
            connection.execute(verifications_helper.update().where(
                verifications_helper.c.id == v.id).values(
                status=new_status))


def downgrade():
    raise exceptions.DowngradeNotSupported()



# Copyright 2016: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from rally import consts
from rally.task import hook


@hook.configure(name="periodic")
class PeriodicTrigger(hook.HookTrigger):
    """Periodically triggers hook with specified range and step."""

    CONFIG_SCHEMA = {
        "type": "object",
        "$schema": consts.JSON_SCHEMA,
        "oneOf": [
            {
                "description": "Periodically triage hook based on elapsed time"
                               " after start of workload.",
                "properties": {
                    "unit": {"enum": ["time"]},
                    "start": {"type": "integer", "minimum": 0},
                    "end": {"type": "integer", "minimum": 1},
                    "step": {"type": "integer", "minimum": 1},
                },
                "required": ["unit", "step"],
                "additionalProperties": False,
            },
            {
                "description": "Periodically triage hook based on iterations.",
                "properties": {
                    "unit": {"enum": ["iteration"]},
                    "start": {"type": "integer", "minimum": 1},
                    "end": {"type": "integer", "minimum": 1},
                    "step": {"type": "integer", "minimum": 1},
                },
                "required": ["unit", "step"],
                "additionalProperties": False,
            },
        ]
    }

    def __init__(self, context, task, hook_cls):
        super(PeriodicTrigger, self).__init__(context, task, hook_cls)
        self.config.setdefault(
            "start", 0 if self.config["unit"] == "time" else 1)
        self.config.setdefault("end", float("Inf"))

    def get_listening_event(self):
        return self.config["unit"]

    def on_event(self, event_type, value=None):
        if not (event_type == self.get_listening_event()
                and self.config["start"] <= value <= self.config["end"]
                and (value - self.config["start"]) % self.config["step"] == 0):
            # do nothing
            return
        super(PeriodicTrigger, self).on_event(event_type, value)



# Copyright 2016: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from rally.common import objects
from rally import exceptions
from tests.unit import test


class VerifierTestCase(test.TestCase):
    def setUp(self):
        super(VerifierTestCase, self).setUp()

        self.db_obj = {"uuid": "uuid-1"}

    def test___str__(self):
        v = objects.Verifier(self.db_obj)
        self.db_obj["name"] = "name"
        self.db_obj["uuid"] = "uuid"
        self.assertEqual("'name' (UUID=uuid)", "%s" % v)

    def test_to_dict(self):
        v = objects.Verifier(self.db_obj)
        v._db_entry = mock.Mock()
        self.assertIsInstance(v.to_dict(), dict)

    @mock.patch("rally.common.objects.verifier.db.verifier_create")
    def test_init(self, mock_verifier_create):
        v = objects.Verifier(self.db_obj)
        self.assertEqual(0, mock_verifier_create.call_count)
        self.assertEqual(self.db_obj["uuid"], v.uuid)
        self.assertEqual(self.db_obj["uuid"], v["uuid"])

    @mock.patch("rally.common.objects.verifier.db.verifier_create")
    def test_create(self, mock_verifier_create):
        objects.Verifier.create("a", "b", "c", "d", "e", False)
        mock_verifier_create.assert_called_once_with(
            name="a", vtype="b", platform="c", source="d", version="e",
            system_wide=False, extra_settings=None)

    @mock.patch("rally.common.objects.verifier.db.verifier_get")
    def test_get(self, mock_verifier_get):
        mock_verifier_get.return_value = self.db_obj
        v = objects.Verifier.get(self.db_obj["uuid"])
        mock_verifier_get.assert_called_once_with(self.db_obj["uuid"])
        self.assertEqual(self.db_obj["uuid"], v.uuid)

    @mock.patch("rally.common.objects.verifier.db.verifier_list")
    def test_list(self, mock_verifier_list):
        mock_verifier_list.return_value = [self.db_obj]
        vs = objects.Verifier.list()
        mock_verifier_list.assert_called_once_with(None)
        self.assertEqual(self.db_obj["uuid"], vs[0].uuid)

    @mock.patch("rally.common.objects.verifier.db.verifier_delete")
    def test_delete(self, mock_verifier_delete):
        objects.Verifier.delete(self.db_obj["uuid"])
        mock_verifier_delete.assert_called_once_with(self.db_obj["uuid"])

    @mock.patch("rally.common.objects.verifier.db.verifier_update")
    def test_update_status(self, mock_verifier_update):
        v = objects.Verifier(self.db_obj)
        v.update_status(status="some-status")
        mock_verifier_update.assert_called_once_with(self.db_obj["uuid"],
                                                     status="some-status")

    @mock.patch("rally.env.env_mgr.EnvManager.get")
    def test_env_and_deployment_properties(self, mock_env_manager_get):
        v = objects.Verifier(self.db_obj)
        v.set_env("some-env")
        self.assertEqual(mock_env_manager_get.return_value,
                         v.deployment.env_obj)
        self.assertEqual(mock_env_manager_get.return_value,
                         v.env)
        mock_env_manager_get.assert_called_once_with("some-env")

    def test_deployment_property_raise_exc(self):
        v = objects.Verifier(self.db_obj)
        self.assertRaises(exceptions.RallyException, getattr, v, "deployment")

    @mock.patch("rally.common.objects.verifier.manager")
    def test_manager_property(self, mock_manager):
        self.db_obj["type"] = "some"
        self.db_obj["platform"] = "platform"
        v = objects.Verifier(self.db_obj)
        self.assertIsNone(v._manager)
        self.assertFalse(mock_manager.VerifierManager.get.called)

        self.assertEqual(
            mock_manager.VerifierManager.get.return_value.return_value,
            v.manager)
        mock_manager.VerifierManager.get.assert_called_once_with(
            self.db_obj["type"], self.db_obj["platform"])



# Copyright (c) 2016 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Merge credentials from users and admin

Revision ID: 3177d36ea270
Revises: ca3626f62937
Create Date: 2016-03-01 16:01:38.747048

"""

from alembic import op
import sqlalchemy as sa

from rally import exceptions


# revision identifiers, used by Alembic.
revision = "3177d36ea270"
down_revision = "ca3626f62937"
branch_labels = None
depends_on = None


deployments_helper = sa.Table(
    "deployments",
    sa.MetaData(),
    sa.Column("id", sa.Integer, primary_key=True, autoincrement=True),
    sa.Column("admin", sa.types.PickleType, nullable=True),
    sa.Column("users", sa.types.PickleType, default=[], nullable=False),
    sa.Column("credentials", sa.types.PickleType, nullable=True),
)


def upgrade():
    with op.batch_alter_table("deployments", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("credentials", sa.PickleType(), nullable=True))

    connection = op.get_bind()
    for deployment in connection.execute(deployments_helper.select()):
        creds = [
            ["openstack",
             {
                 "admin": deployment.admin,
                 "users": deployment.users
             }]
        ]
        connection.execute(
            deployments_helper.update().where(
                deployments_helper.c.id == deployment.id).values(
                credentials=creds))

    with op.batch_alter_table("deployments", schema=None) as batch_op:
        batch_op.alter_column("credentials",
                              existing_type=sa.PickleType,
                              existing_nullable=True,
                              nullable=False)
        batch_op.drop_column("admin")
        batch_op.drop_column("users")


def downgrade():
    raise exceptions.DowngradeNotSupported()



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import print_function

import json
import os
import prettytable
import traceback

from rally.cli import cliutils
from rally.cli import envutils
from rally.common import fileutils
from rally.common import yamlutils as yaml
from rally.env import env_mgr
from rally import exceptions


YES = u":-)"
NO = u":-("


def _print(msg, silent=False):
    if not silent:
        print(msg)


# TODO(boris-42): Wrap all methods to catch EnvManager Exceptions
class EnvCommands(object):
    """Set of commands that allow you to manage envs."""

    @cliutils.args("--name", "-n", type=str, required=True,
                   help="Name of the env.")
    @cliutils.args("--description", "-d", type=str, required=False,
                   help="Env description")
    @cliutils.args("--extras", "-e", type=str, required=False,
                   help="JSON or YAML dict with custom non validate info.")
    @cliutils.args("--from-sysenv", action="store_true", dest="from_sysenv",
                   help="Iterate over all available platforms and check system"
                        " environment for credentials.")
    @cliutils.args("--spec", "-s", type=str, required=False,
                   metavar="<path>", help="Path to env spec.")
    @cliutils.args("--json", action="store_true", dest="to_json",
                   help="Format output as JSON.")
    @cliutils.args("--no-use", action="store_false", dest="do_use",
                   help="Don't set new env as default for future operations.")
    def create(self, api, name, description=None, extras=None,
               spec=None, from_sysenv=False, to_json=False, do_use=True):
        """Create new environment."""

        if spec is not None and from_sysenv:
            print("Arguments '--spec' and '--from-sysenv' cannot be used "
                  "together, use only one of them.")
            return 1
        spec = spec or {}
        if spec:
            with open(os.path.expanduser(spec), "rb") as f:
                spec = yaml.safe_load(f.read())
        if extras:
            extras = yaml.safe_load(extras)

        if from_sysenv:
            result = env_mgr.EnvManager.create_spec_from_sys_environ()
            spec = result["spec"]
            _print("Your system environment includes specifications of "
                   "%s platform(s)." % len(spec), to_json)
            _print("Discovery information:", to_json)
            for p_name, p_result in result["discovery_details"].items():
                _print("\t - %s : %s." % (p_name, p_result["message"]),
                       to_json)

                if "traceback" in p_result:
                    _print("".join(p_result["traceback"]), to_json)
        try:
            env = env_mgr.EnvManager.create(
                name, spec, description=description, extras=extras)
        except exceptions.ManagerInvalidSpec as e:
            _print("Env spec has wrong format:", to_json)
            _print(json.dumps(e.kwargs["spec"], indent=2), to_json)
            for err in e.kwargs["errors"]:
                _print(err, to_json)
            return 1
        except Exception:
            _print("Something went wrong during env creation:", to_json)
            _print(traceback.print_exc(), to_json)
            return 1

        if do_use:
            self._use(env.uuid, to_json)
        self._show(env.data, to_json=to_json, only_spec=False)
        return 0

    @cliutils.args("--json", action="store_true", dest="to_json",
                   help="Format output as JSON.")
    @cliutils.args("--env", dest="env", type=str,
                   metavar="<uuid>", required=False,
                   help="UUID or name of the env.")
    @envutils.with_default_env()
    def cleanup(self, api, env=None, to_json=False):
        """Perform disaster cleanup for specified environment.

        Cases when Rally can leave undeleted resources after performing
        workload:

        - Rally execution was interrupted and cleanup was not performed
        - The environment or a particular platform became unreachable which
          fail Rally execution of cleanup
        """
        env = env_mgr.EnvManager.get(env)
        _print("Cleaning up resources for %s" % env, to_json)
        result = env.cleanup()

        if to_json:
            print(json.dumps(result, indent=2))
            return int(any([p["errors"] for p in result.values()]))

        print("Cleaning is finished. See the results bellow.")

        return_code = 0
        for platform in sorted(result):
            cleanup_info = result[platform]
            print("\nInformation for %s platform." % platform)
            print("=" * 80)
            print("Status: %s" % cleanup_info["message"])
            for key in ("discovered", "deleted", "failed"):
                print("Total %s: %s" % (key, cleanup_info[key]))
            if cleanup_info["errors"]:
                return_code = 1
                errors = "\t- ".join(e["message"]
                                     for e in cleanup_info["errors"])
                print("Errors:\n\t- %s" % errors)

        return return_code

    @cliutils.args("--env", dest="env", type=str,
                   metavar="<uuid>", required=False,
                   help="UUID or name of the env.")
    @cliutils.args("--skip-cleanup", action="store_true", dest="skip_cleanup",
                   help="Do not perform platforms cleanup before destroy.")
    @cliutils.args("--json", action="store_true", dest="to_json",
                   help="Format output as JSON.")
    @cliutils.args("--detailed", action="store_true", dest="detailed",
                   help="Show detailed information.")
    @envutils.with_default_env()
    def destroy(self, api, env=None, skip_cleanup=False, to_json=False,
                detailed=False):
        """Destroy existing environment."""
        env = env_mgr.EnvManager.get(env)
        _print("Destroying %s" % env, to_json)
        result = env.destroy(skip_cleanup)
        return_code = int(result["destroy_info"]["skipped"])

        if result["destroy_info"]["skipped"]:
            _print("%s Failed to destroy env %s: %s"
                   % (NO, env, result["destroy_info"]["message"]), to_json)
        else:
            _print("%s Successfully destroyed env %s" % (YES, env), to_json)

        if detailed or to_json:
            print(json.dumps(result, indent=2))

        return return_code

    @cliutils.args("--env", dest="env", type=str,
                   metavar="<uuid>", required=False,
                   help="UUID or name of the env.")
    @cliutils.args("--force", action="store_true", dest="force",
                   help="Delete DB records even if env is not destroyed.")
    @envutils.with_default_env()
    def delete(self, api, env=None, force=False):
        """Deletes all records related to the environment from db."""
        env_mgr.EnvManager.get(env).delete(force=force)
        # TODO(boris-42): clear env variables if default one is deleted

    MSG_NO_ENVS = ("There are no environments. To create a new environment, "
                   "use command bellow to create one:\nrally env create")

    @cliutils.args("--json", action="store_true", dest="to_json",
                   help="Format output as JSON.")
    @cliutils.suppress_warnings
    def list(self, api, to_json=False):
        """List existing environments."""
        envs = env_mgr.EnvManager.list()
        if to_json:
            print(json.dumps([env.cached_data for env in envs], indent=2))
        elif not envs:
            print(self.MSG_NO_ENVS)
        else:
            cur_env = envutils.get_global(envutils.ENV_ENV)
            table = prettytable.PrettyTable()
            fields = ["uuid", "name", "status", "created_at", "description"]
            table.field_names = fields + ["default"]
            for env in envs:
                row = [env.cached_data[f] for f in fields]
                row.append(cur_env == env.cached_data["uuid"] and "*" or "")
                table.add_row(row)
            table.sortby = "created_at"
            table.reversesort = True
            table.align = "l"
            print(table.get_string())

    def _show(self, env_data, to_json, only_spec):
        if only_spec:
            print(json.dumps(env_data["spec"], indent=2))
        elif to_json:
            print(json.dumps(env_data, indent=2))
        else:
            table = prettytable.PrettyTable()
            table.header = False
            for k in ["uuid", "name", "status",
                      "created_at", "updated_at", "description"]:
                table.add_row([k, env_data[k]])

            table.add_row(["extras", json.dumps(env_data["extras"], indent=2)])
            for p, data in env_data["platforms"].items():
                table.add_row(["platform: %s" % p,
                               json.dumps(data["platform_data"], indent=2)])
            table.align = "l"
            print(table.get_string())

    @cliutils.args("--env", dest="env", type=str,
                   metavar="<uuid>", required=False,
                   help="UUID or name of the env.")
    @cliutils.args("--json", action="store_true", dest="to_json",
                   help="Format output as JSON.")
    @cliutils.args("--only-spec", action="store_true", dest="only_spec",
                   help="Print only a spec for the environment.")
    @cliutils.suppress_warnings
    @envutils.with_default_env()
    def show(self, api, env=None, to_json=False, only_spec=False):
        """Show base information about the environment record."""
        env_data = env_mgr.EnvManager.get(env).data
        self._show(env_data, to_json=to_json, only_spec=only_spec)

    @cliutils.args("--env", dest="env", type=str,
                   metavar="<uuid>", required=False,
                   help="UUID or name of the env.")
    @cliutils.args("--json", action="store_true", dest="to_json",
                   help="Format output as JSON.")
    @envutils.with_default_env()
    def info(self, api, env=None, to_json=False):
        """Retrieve and show environment information."""
        env = env_mgr.EnvManager.get(env)
        env_info = env.get_info()
        return_code = int(any(v.get("error") for v in env_info.values()))

        if to_json:
            print(json.dumps(env_info, indent=2))
            return return_code

        table = prettytable.PrettyTable()
        table.field_names = ["platform", "info", "error"]
        for platform, data in env_info.items():
            table.add_row([
                platform, json.dumps(data["info"], indent=2),
                data.get("error") or ""
            ])
        table.align = "l"
        print(env)
        print(table.get_string())
        return return_code

    @cliutils.args("--env", dest="env", type=str,
                   metavar="<uuid>", required=False,
                   help="UUID or name of the env.")
    @cliutils.args("--json", action="store_true", dest="to_json",
                   help="Format output as JSON.")
    @cliutils.args("--detailed", action="store_true", dest="detailed",
                   help="Show detailed information.")
    @envutils.with_default_env()
    def check(self, api, env=None, to_json=False, detailed=False):
        """Check availability of all platforms in environment."""
        env = env_mgr.EnvManager.get(env)
        data = env.check_health()
        available = all(x["available"] for x in data.values())

        if to_json:
            print(json.dumps(data, indent=2))
            return not available

        def _format_raw(plugin_name, el):
            return [
                el["available"] and YES or NO,
                plugin_name.split("@")[1], el["message"], plugin_name
            ]

        table = prettytable.PrettyTable()
        if detailed:
            table.field_names = ["Available", "Platform", "Message", "Plugin"]
            for plugin_name, r in data.items():
                table.add_row(_format_raw(plugin_name, r))
        else:
            table.field_names = ["Available", "Platform", "Message"]
            for plugin_name, r in data.items():
                table.add_row(_format_raw(plugin_name, r)[:3])

        table.align = "l"
        table.align["available"] = "c"
        table.sortby = "Platform"
        print("%s %s" % (env, available and YES or NO))
        print(table.get_string())
        if not available and detailed:
            for name, p_data in data.items():
                if p_data["available"]:
                    continue
                print("-" * 4)
                print("Plugin %s raised exception:" % name)
                print("".join(p_data["traceback"]))

        return not available

    @cliutils.args("--env", dest="env", type=str,
                   metavar="<uuid>", required=False,
                   help="UUID or name of a env.")
    @cliutils.args("--json", action="store_true", dest="to_json",
                   help="Format output as JSON.")
    def use(self, api, env, to_json=False):
        """Set default environment."""
        try:
            env = env_mgr.EnvManager.get(env)
        except exceptions.DBRecordNotFound:
            _print("Can't use non existing environment %s." % env, to_json)
            return 1
        self._use(env.uuid, to_json)

    def _use(self, env_uuid, to_json):
        _print("Using environment: %s" % env_uuid, to_json)
        fileutils.update_globals_file(envutils.ENV_ENV, env_uuid)



# Copyright 2016: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import subprocess

import testtools

from tests.functional import utils


class LibAPITestCase(testtools.TestCase):

    def test_rally_lib(self):
        rally = utils.Rally(force_new_db=True)
        cdir = os.path.dirname(os.path.realpath(__file__))
        app = os.path.join(cdir, "../ci/rally_app.py")
        subprocess.check_output(
            ["python", app, "--config-file", rally.config_filename])



# Copyright 2017: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt

from rally.common.plugin import plugin
from rally.common import validation
from rally import exceptions
from tests.unit import test


class ValidationHelpersTestCase(test.TestCase):

    def test_validation_add_validator_to_validator(self):
        exc = self.assertRaises(
            exceptions.RallyException,
            validation.add("dummy_validator"), DummyValidator)
        self.assertEqual("Only RequiredPlatformValidator can be added "
                         "to other validators as a validator", str(exc))

    def test_validation_add_validator_to_platform_validator(self):
        exc = self.assertRaises(
            exceptions.RallyException,
            validation.add("required_platform"),
            validation.RequiredPlatformValidator)
        self.assertEqual("Cannot add a validator to RequiredPlatformValidator",
                         str(exc))


@validation.add(name="required_platform", platform="foo", admin=True)
@plugin.configure(name="dummy_validator")
class DummyValidator(validation.Validator):

    def __init__(self, foo, exc=False):
        """Dummy validator

        :param foo: additional parameter for Dummy validator
        :param exc: whether to raise expected on unexpected error
        """
        super(DummyValidator, self).__init__()
        self.foo = foo
        self.exc = exc

    def validate(self, context, config, plugin_cls, plugin_cfg):
        if self.foo not in config:
            if self.exc:
                raise Exception("foo")
            self.fail("oops")


@plugin.base()
class DummyPluginBase(plugin.Plugin,
                      validation.ValidatablePluginMixin):
    pass


class ValidatorTestCase(test.TestCase):

    def test_dummy_validators(self):
        @plugin.base()
        class DummyPluginBase(plugin.Plugin,
                              validation.ValidatablePluginMixin):
            pass

        @validation.add(name="dummy_validator", foo="bar", exc=True)
        @validation.add(name="required_platform", platform="foo", users=True)
        @plugin.configure(name="dummy_plugin")
        class DummyPlugin(DummyPluginBase):
            pass

        ctx = {"platforms": {
            "foo": {"admin": "fake_admin", "users": ["fake_user"]}}}
        result = DummyPluginBase.validate(
            name="dummy_plugin", context=ctx,
            config={"bar": 1}, plugin_cfg={})
        self.assertEqual(0, len(result), result)

        result = DummyPluginBase.validate(
            name="dummy_plugin", context=ctx, config={}, plugin_cfg={})
        self.assertEqual(1, len(result))
        self.assertIn("raise Exception(\"foo\")", result[0])

        DummyPlugin.unregister()

    def test_simple_failure(self):

        result = DummyPluginBase.validate("dummy_plugin", None, None, None)
        self.assertEqual(1, len(result))
        self.assertIn("There is no DummyPluginBase plugin `dummy_plugin`",
                      result[0])

    def test_failure_includes_detailed_info(self):

        @validation.add("dummy_validator", foo="bar")
        @plugin.configure(name=self.id())
        class Foo(DummyPluginBase):
            pass

        result = DummyPluginBase.validate(self.id(), {}, {}, None,
                                          vtype="semantic")
        self.assertEqual(1, len(result))
        self.assertEqual(
            "DummyPluginBase plugin '%s' doesn't pass dummy_validator@default "
            "validation. Details: oops" % self.id(),
            result[0])


@ddt.ddt
class RequiredPlatformValidatorTestCase(test.TestCase):

    @ddt.data(
        {"kwargs": {"platform": "foo"},
         "context": {"platforms": {"foo": {}}}},
        {"kwargs": {"platform": "openstack", "admin": True},
         "context": {"platforms": {"openstack": {"admin": "fake_admin"}}}},
        {"kwargs": {"platform": "openstack", "admin": True, "users": True},
         "context": {"platforms": {"openstack": {"admin": "fake_admin"}}}},
        {"kwargs": {"platform": "openstack", "admin": True, "users": True},
         "context": {"platforms": {"openstack": {"admin": "fake_admin",
                                                 "users": ["fake_user"]}}}}
    )
    @ddt.unpack
    def test_validator(self, kwargs, context):
        validator = validation.RequiredPlatformValidator(**kwargs)
        validator.validate(context, None, None, None)

    @ddt.data(
        {"kwargs": {"platform": "openstack", "admin": True},
         "context": {"platforms": {"openstack": {}}},
         "error_msg": "No admin credential for openstack"},
        {"kwargs": {"platform": "openstack", "users": True},
         "context": {"platforms": {"openstack": {}}},
         "error_msg": "No user credentials for openstack"},
        {"kwargs": {"platform": "openstack"},
         "context": {"platforms": {"openstack": {}}},
         "error_msg": "You should specify admin=True or users=True or both "
                      "for validating openstack platform."},
        {"kwargs": {"platform": "foo"},
         "context": {"platforms": {}},
         "error_msg": "There is no specification for foo platform in "
                      "selected environment."}
    )
    @ddt.unpack
    def test_validator_failed(self, kwargs, context, error_msg):
        validator = validation.RequiredPlatformValidator(**kwargs)
        e = self.assertRaises(
            validation.ValidationError,
            validator.validate, context, None, None, None)
        self.assertEqual(error_msg, e.message)



# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

import ddt
import mock

from rally.utils import strutils
from tests.unit import test


@ddt.ddt
class StrUtilsTestCase(test.TestCase):

    def test_is_uuid_like(self):
        self.assertTrue(strutils.is_uuid_like(str(uuid.uuid4())))
        self.assertTrue(strutils.is_uuid_like(
            "{12345678-1234-5678-1234-567812345678}"))
        self.assertTrue(strutils.is_uuid_like(
            "12345678123456781234567812345678"))
        self.assertTrue(strutils.is_uuid_like(
            "urn:uuid:12345678-1234-5678-1234-567812345678"))
        self.assertTrue(strutils.is_uuid_like(
            "urn:bbbaaaaa-aaaa-aaaa-aabb-bbbbbbbbbbbb"))
        self.assertTrue(strutils.is_uuid_like(
            "uuid:bbbaaaaa-aaaa-aaaa-aabb-bbbbbbbbbbbb"))
        self.assertTrue(strutils.is_uuid_like(
            "{}---bbb---aaa--aaa--aaa-----aaa---aaa--bbb-bbb---bbb-bbb-bb-{}"))

    def test_is_uuid_like_insensitive(self):
        self.assertTrue(strutils.is_uuid_like(str(uuid.uuid4()).upper()))

    def test_id_is_uuid_like(self):
        self.assertFalse(strutils.is_uuid_like(1234567))

    def test_name_is_uuid_like(self):
        self.assertFalse(strutils.is_uuid_like("asdasdasd"))

    @mock.patch("builtins.str")
    def test_bool_bool_from_string_no_text(self, mock_str):
        self.assertTrue(strutils.bool_from_string(True))
        self.assertFalse(strutils.bool_from_string(False))
        self.assertEqual(0, mock_str.call_count)

    def test_bool_bool_from_string(self):
        self.assertTrue(strutils.bool_from_string(True))
        self.assertFalse(strutils.bool_from_string(False))

    def test_bool_bool_from_string_default(self):
        self.assertTrue(strutils.bool_from_string("", default=True))
        self.assertFalse(strutils.bool_from_string("wibble", default=False))

    def _test_bool_from_string(self, c):
        self.assertTrue(strutils.bool_from_string(c("true")))
        self.assertTrue(strutils.bool_from_string(c("TRUE")))
        self.assertTrue(strutils.bool_from_string(c("on")))
        self.assertTrue(strutils.bool_from_string(c("On")))
        self.assertTrue(strutils.bool_from_string(c("yes")))
        self.assertTrue(strutils.bool_from_string(c("YES")))
        self.assertTrue(strutils.bool_from_string(c("yEs")))
        self.assertTrue(strutils.bool_from_string(c("1")))
        self.assertTrue(strutils.bool_from_string(c("T")))
        self.assertTrue(strutils.bool_from_string(c("t")))
        self.assertTrue(strutils.bool_from_string(c("Y")))
        self.assertTrue(strutils.bool_from_string(c("y")))

        self.assertFalse(strutils.bool_from_string(c("false")))
        self.assertFalse(strutils.bool_from_string(c("FALSE")))
        self.assertFalse(strutils.bool_from_string(c("off")))
        self.assertFalse(strutils.bool_from_string(c("OFF")))
        self.assertFalse(strutils.bool_from_string(c("no")))
        self.assertFalse(strutils.bool_from_string(c("0")))
        self.assertFalse(strutils.bool_from_string(c("42")))
        self.assertFalse(strutils.bool_from_string(c(
                         "This should not be True")))
        self.assertFalse(strutils.bool_from_string(c("F")))
        self.assertFalse(strutils.bool_from_string(c("f")))
        self.assertFalse(strutils.bool_from_string(c("N")))
        self.assertFalse(strutils.bool_from_string(c("n")))

        # Whitespace should be stripped
        self.assertTrue(strutils.bool_from_string(c(" 1 ")))
        self.assertTrue(strutils.bool_from_string(c(" true ")))
        self.assertFalse(strutils.bool_from_string(c(" 0 ")))
        self.assertFalse(strutils.bool_from_string(c(" false ")))

    def test_bool_from_string(self):
        self._test_bool_from_string(lambda s: s)

    def test_other_bool_from_string(self):
        self.assertFalse(strutils.bool_from_string(None))
        self.assertFalse(strutils.bool_from_string(mock.Mock()))

    def test_int_bool_from_string(self):
        self.assertTrue(strutils.bool_from_string(1))

        self.assertFalse(strutils.bool_from_string(-1))
        self.assertFalse(strutils.bool_from_string(0))
        self.assertFalse(strutils.bool_from_string(2))

    def test_strict_bool_from_string(self):
        # None isn"t allowed in strict mode
        exc = self.assertRaises(ValueError, strutils.bool_from_string, None,
                                strict=True)
        expected_msg = ("Unrecognized value 'None', acceptable values are:"
                        " '0', '1', 'f', 'false', 'n', 'no', 'off', 'on',"
                        " 't', 'true', 'y', 'yes'")
        self.assertEqual(expected_msg, str(exc))

        # Unrecognized strings aren't allowed
        self.assertFalse(strutils.bool_from_string("Other", strict=False))
        exc = self.assertRaises(ValueError, strutils.bool_from_string, "Other",
                                strict=True)
        expected_msg = ("Unrecognized value 'Other', acceptable values are:"
                        " '0', '1', 'f', 'false', 'n', 'no', 'off', 'on',"
                        " 't', 'true', 'y', 'yes'")
        self.assertEqual(expected_msg, str(exc))

        # Unrecognized numbers aren't allowed
        exc = self.assertRaises(ValueError, strutils.bool_from_string, 2,
                                strict=True)
        expected_msg = ("Unrecognized value '2', acceptable values are:"
                        " '0', '1', 'f', 'false', 'n', 'no', 'off', 'on',"
                        " 't', 'true', 'y', 'yes'")
        self.assertEqual(expected_msg, str(exc))

        # False-like values are allowed
        self.assertFalse(strutils.bool_from_string("f", strict=True))
        self.assertFalse(strutils.bool_from_string("false", strict=True))
        self.assertFalse(strutils.bool_from_string("off", strict=True))
        self.assertFalse(strutils.bool_from_string("n", strict=True))
        self.assertFalse(strutils.bool_from_string("no", strict=True))
        self.assertFalse(strutils.bool_from_string("0", strict=True))

        self.assertTrue(strutils.bool_from_string("1", strict=True))

        # Avoid font-similarity issues (one looks like lowercase-el, zero like
        # oh, etc...)
        for char in ("O", "o", "L", "l", "I", "i"):
            self.assertRaises(ValueError, strutils.bool_from_string, char,
                              strict=True)

    @ddt.data(
        {
            "num_float": 0,
            "num_str": "0.0"
        },
        {
            "num_float": 37,
            "num_str": "37.0"
        },
        {
            "num_float": 0.0000001,
            "num_str": "0.0"
        },
        {
            "num_float": 0.000000,
            "num_str": "0.0"
        },
        {
            "num_float": 1.0000001,
            "num_str": "1.0"
        },
        {
            "num_float": 1.0000011,
            "num_str": "1.000001"
        },
        {
            "num_float": 1.0000019,
            "num_str": "1.000002"
        }

    )
    @ddt.unpack
    def test_format_float_to_str(self, num_float, num_str):
        self.assertEqual(num_str, strutils.format_float_to_str(num_float))



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Add Env & Platforms tables

Revision ID: a43700a813a5
Revises: dc46687661df
Create Date: 2017-12-27 13:37:10.144970

"""

from alembic import op
import sqlalchemy as sa

from rally.common.db import sa_types
from rally import exceptions


# revision identifiers, used by Alembic.
revision = "a43700a813a5"
down_revision = "44169f4d455e"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "envs",
        sa.Column("id", sa.Integer, primary_key=True, autoincrement=True),
        sa.Column("uuid", sa.String(36), nullable=False),

        sa.Column("name", sa.String(255), nullable=False),
        sa.Column("description", sa.Text, default=""),
        sa.Column("status", sa.String(36), nullable=False),

        sa.Column("extras", sa_types.MutableJSONEncodedDict, default={}),
        sa.Column("spec", sa_types.MutableJSONEncodedDict, default={}),

        sa.Column("created_at", sa.DateTime),
        sa.Column("updated_at", sa.DateTime)
    )

    op.create_index("env_uuid", "envs", ["uuid"], unique=True)
    op.create_index("env_name", "envs", ["name"], unique=True)
    op.create_index("env_status", "envs", ["status"])

    op.create_table(
        "platforms",
        sa.Column("id", sa.Integer, primary_key=True, autoincrement=True),
        sa.Column("uuid", sa.String(36), nullable=False),
        sa.Column("env_uuid", sa.String(36), nullable=False),

        sa.Column("status", sa.String(36), nullable=False),

        sa.Column("plugin_name", sa.String(36), nullable=False),
        sa.Column("plugin_spec", sa_types.MutableJSONEncodedDict,
                  nullable=False),
        sa.Column("plugin_data", sa_types.MutableJSONEncodedDict,
                  default={}),

        sa.Column("platform_name", sa.String(36)),
        sa.Column("platform_data", sa_types.MutableJSONEncodedDict,
                  default={}),

        sa.Column("created_at", sa.DateTime),
        sa.Column("updated_at", sa.DateTime),

    )

    op.create_index("platform_uuid", "platforms", ["uuid"], unique=True)
    op.create_index("platform_env_uuid", "platforms", ["env_uuid"])


def downgrade():
    raise exceptions.DowngradeNotSupported()



# Copyright 2014: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


"""
SLA (Service-level agreement) is set of details for determining compliance
with contracted values such as maximum error rate or minimum response time.
"""

from rally import consts
from rally.task import sla


@sla.configure(name="failure_rate")
class FailureRate(sla.SLA):
    """Failure rate minimum and maximum in percents."""
    CONFIG_SCHEMA = {
        "type": "object",
        "$schema": consts.JSON_SCHEMA,
        "properties": {
            "min": {"type": "number", "minimum": 0.0, "maximum": 100.0},
            "max": {"type": "number", "minimum": 0.0, "maximum": 100.0}
        },
        "minProperties": 1,
        "additionalProperties": False,
    }

    def __init__(self, criterion_value):
        super(FailureRate, self).__init__(criterion_value)
        self.min_percent = self.criterion_value.get("min", 0)
        self.max_percent = self.criterion_value.get("max", 100)
        self.errors = 0
        self.total = 0
        self.error_rate = 0.0

    def add_iteration(self, iteration):
        self.total += 1
        if iteration["error"]:
            self.errors += 1
        self.error_rate = self.errors * 100.0 / self.total
        self.success = self.min_percent <= self.error_rate <= self.max_percent
        return self.success

    def merge(self, other):
        self.total += other.total
        self.errors += other.errors
        if self.total:
            self.error_rate = self.errors * 100.0 / self.total
        self.success = self.min_percent <= self.error_rate <= self.max_percent
        return self.success

    def details(self):
        return ("Failure rate criteria %.2f%% <= %.2f%% <= %.2f%% - %s" %
                (self.min_percent, self.error_rate,
                 self.max_percent, self.status()))



# Copyright 2015: eNovance
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import xml.etree.ElementTree as ET


class JUnit(object):
    SUCCESS = "success"
    FAILURE = "failure"
    ERROR = "error"

    def __init__(self, test_suite_name):
        self.test_suite_name = test_suite_name
        self.test_cases = []
        self.n_tests = 0
        self.n_failures = 0
        self.n_errors = 0
        self.total_time = 0.0

    def add_test(self, test_name, time, outcome=SUCCESS, message=""):
        class_name, name = test_name.split(".", 1)
        self.test_cases.append({
            "classname": class_name,
            "name": name,
            "time": str("%.2f" % time),
            "outcome": outcome,
            "message": message
        })

        if outcome == JUnit.FAILURE:
            self.n_failures += 1
        elif outcome == JUnit.ERROR:
            self.n_errors += 1
        elif outcome != JUnit.SUCCESS:
            raise ValueError("Unexpected outcome %s" % outcome)

        self.n_tests += 1
        self.total_time += time

    def to_xml(self):
        xml = ET.Element("testsuite", {
            "name": self.test_suite_name,
            "tests": str(self.n_tests),
            "time": str("%.2f" % self.total_time),
            "failures": str(self.n_failures),
            "errors": str(self.n_errors),
        })
        for test_case in self.test_cases:
            outcome = test_case.pop("outcome")
            message = test_case.pop("message")
            if outcome in [JUnit.FAILURE, JUnit.ERROR]:
                sub = ET.SubElement(xml, "testcase", test_case)
                sub.append(ET.Element(outcome, {"message": message}))
            else:
                xml.append(ET.Element("testcase", test_case))
        return ET.tostring(xml, encoding="utf-8").decode("utf-8")



# Copyright 2015: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys

from rally.common.plugin import discover
from rally.common.plugin import info
from rally.common.plugin import meta
from rally import exceptions


def base():
    """Mark Plugin as a base.

    Base Plugins are used to have better organization of plugins providing
    subtypes. Base Plugins may contain documentation and other base methods
    related to specific type of plugins, resolving next problems:

    - Having different types of plugins (e.g. Scenario, Context, SLA, ...)
    - Auto generation of plugin reference with splitting plugins by their base
    - Plugin lookup - one can easily get all plugins from some base.

    Plugin bases by default initialize _default_meta
    """
    def wrapper(cls):
        if not issubclass(cls, Plugin):
            raise exceptions.RallyException(
                "Plugin's Base can be only a subclass of Plugin class.")

        parent = cls._get_base()
        if parent != Plugin:
            raise exceptions.RallyException(
                "'%(plugin_cls)s' can not be marked as plugin base, since it "
                "inherits from '%(parent)s' which is also plugin base."
                % {"plugin_cls": cls.__name__, "parent": parent.__name__})

        cls.base_ref = cls
        cls._default_meta_init(True)
        return cls
    return wrapper


def configure(name, platform="default", hidden=False):
    """Use this decorator to configure plugin's attributes.

    Plugin is not discoverable until configure() is performed.

    :param name: name of plugin that is used for searching purpose
    :param platform: platform name that plugin belongs to
    :param hidden: if True the plugin will be marked as hidden and can be
        loaded only explicitly
    """

    def decorator(plugin):
        plugin_id = "%s.%s" % (plugin.__module__, plugin.__name__)
        if not name:
            raise ValueError(
                "The name of the plugin %s cannot be empty." % plugin_id)
        if "@" in name:
            raise ValueError(
                "The name of the plugin cannot contain @ symbol" % plugin_id)

        plugin._meta_init()
        try:
            existing_plugin = plugin._get_base().get(
                name=name, platform=platform, allow_hidden=True)
        except exceptions.PluginNotFound:
            plugin._meta_set("name", name)
            plugin._meta_set("platform", platform)
        else:
            plugin.unregister()
            raise exceptions.PluginWithSuchNameExists(
                name=name, platform=existing_plugin.get_platform(),
                existing_path=(
                    sys.modules[existing_plugin.__module__].__file__),
                new_path=sys.modules[plugin.__module__].__file__
            )
        plugin._meta_set("hidden", hidden)
        return plugin

    return decorator


def default_meta(inherit=True):
    """Initialize default meta for particular plugin.

    Default Meta is inherited by all children comparing to Meta which is unique
    per plugin.

    :param inherit: Whatever to copy parents default meta
    """

    def decorator(plugin):
        plugin._default_meta_init(inherit)
        return plugin

    return decorator


def deprecated(reason, rally_version):
    """Mark plugin as deprecated.

    :param reason: Message that describes reason of plugin deprecation
    :param rally_version: Deprecated since this version of Rally
    """
    def decorator(plugin):
        plugin._meta_set("deprecated", {
            "reason": reason,
            "rally_version": rally_version
        })
        return plugin

    return decorator


class Plugin(meta.MetaMixin, info.InfoMixin):
    """Base class for all Plugins in Rally."""

    @classmethod
    def unregister(cls):
        """Removes all plugin meta information and makes it undiscoverable."""
        cls._meta_clear()

    @classmethod
    def _get_base(cls):
        return getattr(cls, "base_ref", Plugin)

    @classmethod
    def get(cls, name, platform=None, allow_hidden=False):
        """Return plugin by its name for specified platform.

        :param name: Plugin's name or fullname
        :param platform: Plugin's platform
        :param allow_hidden: if False and found plugin is hidden then
                              PluginNotFound is raised
        """
        if "@" in name:
            name, platform = name.split("@", 1)

        results = cls.get_all(name=name, platform=platform,
                              allow_hidden=allow_hidden)

        if not results:
            base = cls._get_base()
            base = "" if base == Plugin else " %s" % base.__name__
            raise exceptions.PluginNotFound(
                name=name, platform=platform or "any", base=base)

        if len(results) == 1:
            return results[0]

        if platform is None:
            # try to use default platform
            default = [p for p in results if p.get_platform() == "default"]
            if default:
                return default[0]

        raise exceptions.MultiplePluginsFound(
            name=name,
            plugins=", ".join(p.get_fullname() for p in results))

    @classmethod
    def get_all(cls, platform=None, allow_hidden=False, name=None):
        """Return all subclass plugins of plugin.

        All plugins that are not configured will be ignored.
        :param platform: return only plugins for specific platform.
        :param name: return only plugins with specified name.
        :param allow_hidden: if False return only non hidden plugins
        """
        plugins = []

        for p in discover.itersubclasses(cls):
            if not issubclass(p, Plugin):
                continue
            if not p._meta_is_inited(raise_exc=False):
                continue
            if name and name != p.get_name():
                continue
            if platform and platform != p.get_platform():
                continue
            if not allow_hidden and p.is_hidden():
                continue
            plugins.append(p)

        return plugins

    @classmethod
    def get_name(cls):
        """Return plugin's name."""
        return cls._meta_get("name")

    @classmethod
    def get_platform(cls):
        """"Return plugin's platform name."""
        return cls._meta_get("platform")

    @classmethod
    def get_fullname(cls):
        """Returns plugins's full name."""
        return "%s@%s" % (cls.get_name(), cls.get_platform() or "")

    @classmethod
    def is_hidden(cls):
        """Returns whatever plugin is hidden or not."""
        return cls._meta_get("hidden", False)

    @classmethod
    def is_deprecated(cls):
        """Returns deprecation details if plugin is deprecated."""
        return cls._meta_get("deprecated", False)



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

import requests

from rally.common import logging
from rally import exceptions

LOG = logging.getLogger(__name__)


class ElasticSearchClient(object):
    """The helper class for communication with ElasticSearch 2.*, 5.*, 6.*"""

    # a number of documents to push to the cluster at once.
    CHUNK_LENGTH = 10000

    def __init__(self, url):
        self._url = url.rstrip("/") if url else "http://localhost:9200"
        self._version = None

    @staticmethod
    def _check_response(resp, action=None):
        if resp.status_code in (200, 201):
            return
        # it is an error. let's try to find the reason
        reason = None
        try:
            data = resp.json()
        except ValueError:
            # it is ok
            pass
        else:
            if "error" in data:
                if isinstance(data["error"], dict):
                    reason = data["error"].get("reason", "")
                else:
                    reason = data["error"]
        reason = reason or resp.text or "n/a"
        action = action or "connect to"
        raise exceptions.RallyException(
            "[HTTP %s] Failed to %s ElasticSearch cluster: %s" %
            (resp.status_code, action, reason))

    def version(self):
        """Get version of the ElasticSearch cluster."""
        if self._version is None:
            self.info()
        return self._version

    def info(self):
        """Retrieve info about the ElasticSearch cluster."""
        resp = requests.get(self._url)
        self._check_response(resp)
        err_msg = "Failed to retrieve info about the ElasticSearch cluster: %s"
        try:
            data = resp.json()
        except ValueError:
            LOG.debug("Return data from %s: %s" % (self._url, resp.text))
            raise exceptions.RallyException(
                err_msg % "The return data doesn't look like a json.")
        version = data.get("version", {}).get("number")
        if not version:
            LOG.debug("Return data from %s: %s" % (self._url, resp.text))
            raise exceptions.RallyException(
                err_msg % "Failed to parse the received data.")
        self._version = version
        if self._version.startswith("2"):
            data["version"]["build_date"] = data["version"].pop(
                "build_timestamp")
        return data

    def push_documents(self, documents):
        """Push documents to the ElasticSearch cluster using bulk API.

        :param documents: a list of documents to push
        """
        LOG.debug("Pushing %s documents by chunks (up to %s documents at once)"
                  " to ElasticSearch." %
                  # dividing numbers by two, since each documents has 2 lines
                  #     in `documents` (action and document itself).
                  (len(documents) / 2, self.CHUNK_LENGTH / 2))

        for pos in range(0, len(documents), self.CHUNK_LENGTH):
            data = "\n".join(documents[pos:pos + self.CHUNK_LENGTH]) + "\n"

            raw_resp = requests.post(
                self._url + "/_bulk", data=data,
                headers={"Content-Type": "application/x-ndjson"}
            )
            self._check_response(raw_resp, action="push documents to")

            LOG.debug("Successfully pushed %s documents." %
                      len(raw_resp.json()["items"]))

    def list_indices(self):
        """List all indices."""
        resp = requests.get(self._url + "/_cat/indices?v")
        self._check_response(resp, "list the indices at")

        return resp.text.rstrip().split(" ")

    def create_index(self, name, doc_type, properties):
        """Create an index.

        There are two very different ways to search strings. You can either
        search whole values, that we often refer to as keyword search, or
        individual tokens, that we usually refer to as full-text search.
        In ElasticSearch 2.x `string` data type is used for these cases whereas
        ElasticSearch 5.0 the `string` data type was replaced by two new types:
        `keyword` and `text`. Since it is hard to predict the destiny of
        `string` data type and support of 2 formats of input data, the
        properties should be transmitted in ElasticSearch 5.x format.
        """
        if self.version().startswith("2."):
            properties = copy.deepcopy(properties)
            for spec in properties.values():
                if spec.get("type", None) == "text":
                    spec["type"] = "string"
                elif spec.get("type", None) == "keyword":
                    spec["type"] = "string"
                    spec["index"] = "not_analyzed"

        resp = requests.put(
            self._url + "/%s" % name,
            json={"mappings": {doc_type: {"properties": properties}}})
        self._check_response(resp, "create index at")

    def check_document(self, index, doc_id, doc_type="data"):
        """Check for the existence of a document.

        :param index: The index of a document
        :param doc_id: The ID of a document
        :param doc_type: The type of a document (Defaults to data)
        """
        resp = requests.head("%(url)s/%(index)s/%(type)s/%(id)s" %
                             {"url": self._url,
                              "index": index,
                              "type": doc_type,
                              "id": doc_id})
        if resp.status_code == 200:
            return True
        elif resp.status_code == 404:
            return False
        else:
            self._check_response(resp, "check the index at")



# Copyright 2014: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import multiprocessing
import queue as Queue
import threading
import time


from rally.common import logging
from rally.common import utils
from rally.common import validation
from rally import consts
from rally.task import runner

LOG = logging.getLogger(__name__)


def _worker_process(queue, iteration_gen, timeout, times, max_concurrent,
                    context, cls, method_name, args, event_queue, aborted,
                    runs_per_second, rps_cfg, processes_to_start, info):
    """Start scenario within threads.

    Spawn N threads per second. Each thread runs the scenario once, and appends
    result to queue. A maximum of max_concurrent threads will be ran
    concurrently.

    :param queue: queue object to append results
    :param iteration_gen: next iteration number generator
    :param timeout: operation's timeout
    :param times: total number of scenario iterations to be run
    :param max_concurrent: maximum worker concurrency
    :param context: scenario context object
    :param cls: scenario class
    :param method_name: scenario method name
    :param args: scenario args
    :param aborted: multiprocessing.Event that aborts load generation if
                    the flag is set
    :param runs_per_second: function that should return desired rps value
    :param rps_cfg: rps section from task config
    :param processes_to_start: int, number of started processes for scenario
                               execution
    :param info: info about all processes count and counter of runned process
    """

    pool = collections.deque()
    if isinstance(rps_cfg, dict):
        rps = rps_cfg["start"]
    else:
        rps = rps_cfg
    sleep = 1.0 / rps

    runner._log_worker_info(times=times, rps=rps, timeout=timeout,
                            cls=cls, method_name=method_name, args=args)

    time.sleep(
        (sleep * info["processes_counter"]) / info["processes_to_start"])

    start = time.time()
    timeout_queue = Queue.Queue()

    if timeout:
        collector_thr_by_timeout = threading.Thread(
            target=utils.timeout_thread,
            args=(timeout_queue, )
        )
        collector_thr_by_timeout.start()

    i = 0
    while i < times and not aborted.is_set():
        scenario_context = runner._get_scenario_context(next(iteration_gen),
                                                        context)
        worker_args = (
            queue, cls, method_name, scenario_context, args, event_queue)
        thread = threading.Thread(target=runner._worker_thread,
                                  args=worker_args)

        i += 1
        thread.start()
        if timeout:
            timeout_queue.put((thread, time.time() + timeout))
        pool.append(thread)

        time_gap = time.time() - start
        real_rps = i / time_gap if time_gap else "Infinity"

        LOG.debug(
            "Worker: %s rps: %s (requested rps: %s)" %
            (i, real_rps, runs_per_second(rps_cfg, start, processes_to_start)))

        # try to join latest thread(s) until it finished, or until time to
        # start new thread (if we have concurrent slots available)
        while i / (time.time() - start) > runs_per_second(
                rps_cfg, start, processes_to_start) or (
                    len(pool) >= max_concurrent):
            if pool:
                pool[0].join(0.001)
                if not pool[0].is_alive():
                    pool.popleft()
            else:
                time.sleep(0.001)

    while pool:
        pool.popleft().join()

    if timeout:
        timeout_queue.put((None, None,))
        collector_thr_by_timeout.join()


@validation.configure("check_rps")
class CheckPRSValidator(validation.Validator):
    """Additional schema validation for rps runner"""

    def validate(self, context, config, plugin_cls, plugin_cfg):
        if isinstance(plugin_cfg["rps"], dict):
            if plugin_cfg["rps"]["end"] < plugin_cfg["rps"]["start"]:
                msg = "rps end value must not be less than rps start value."
                return self.fail(msg)


@validation.add("check_rps")
@runner.configure(name="rps")
class RPSScenarioRunner(runner.ScenarioRunner):
    """Scenario runner that does the job with specified frequency.

    Every single scenario iteration is executed with specified frequency
    (runs per second) in a pool of processes. The scenario will be
    launched for a fixed number of times in total (specified in the config).

    An example of a rps scenario is booting 1 VM per second. This
    execution type is thus very helpful in understanding the maximal load that
    a certain cloud can handle.
    """

    CONFIG_SCHEMA = {
        "type": "object",
        "$schema": consts.JSON_SCHEMA7,
        "properties": {
            "times": {
                "type": "integer",
                "minimum": 1
            },
            "rps": {
                "anyOf": [
                    {
                        "description": "Generate constant requests per second "
                                       "during the whole workload.",
                        "type": "number",
                        "exclusiveMinimum": 0,
                        "minimum": 0
                    },
                    {
                        "type": "object",
                        "description": "Increase requests per second for "
                                       "specified value each time after a "
                                       "certain number of seconds.",
                        "properties": {
                            "start": {
                                "type": "number",
                                "minimum": 1
                            },
                            "end": {
                                "type": "number",
                                "minimum": 1
                            },
                            "step": {
                                "type": "number",
                                "minimum": 1
                            },
                            "duration": {
                                "type": "number",
                                "minimum": 1
                            }
                        },
                        "additionalProperties": False,
                        "required": ["start", "end", "step"]
                    }
                ],
            },
            "timeout": {
                "type": "number",
            },
            "max_concurrency": {
                "type": "integer",
                "minimum": 1
            },
            "max_cpu_count": {
                "type": "integer",
                "minimum": 1
            }
        },
        "required": ["times", "rps"],
        "additionalProperties": False
    }

    def _run_scenario(self, cls, method_name, context, args):
        """Runs the specified scenario with given arguments.

        Every single scenario iteration is executed with specified
        frequency (runs per second) in a pool of processes. The scenario is
        launched for a fixed number of times in total (specified in the
        config).

        :param cls: The Scenario class where the scenario is implemented
        :param method_name: Name of the method that implements the scenario
        :param context: Context that contains users, admin & other
                        information, that was created before scenario
                        execution starts.
        :param args: Arguments to call the scenario method with

        :returns: List of results fore each single scenario iteration,
                  where each result is a dictionary
        """
        times = self.config["times"]
        timeout = self.config.get("timeout", 0)  # 0 means no timeout
        iteration_gen = utils.RAMInt()

        cpu_count = multiprocessing.cpu_count()
        max_cpu_used = min(cpu_count,
                           self.config.get("max_cpu_count", cpu_count))

        def runs_per_second(rps_cfg, start_timer, number_of_processes):
            """At the given second return desired rps."""

            if not isinstance(rps_cfg, dict):
                return float(rps_cfg) / number_of_processes
            stage_order = (time.time() - start_timer) / rps_cfg.get(
                "duration", 1) - 1
            rps = (float(rps_cfg["start"] + rps_cfg["step"] * stage_order)
                   / number_of_processes)

            return min(rps, float(rps_cfg["end"]))

        processes_to_start = min(max_cpu_used, times,
                                 self.config.get("max_concurrency", times))
        times_per_worker, times_overhead = divmod(times, processes_to_start)

        # Determine concurrency per worker
        concurrency_per_worker, concurrency_overhead = divmod(
            self.config.get("max_concurrency", times), processes_to_start)

        self._log_debug_info(times=times, timeout=timeout,
                             max_cpu_used=max_cpu_used,
                             processes_to_start=processes_to_start,
                             times_per_worker=times_per_worker,
                             times_overhead=times_overhead,
                             concurrency_per_worker=concurrency_per_worker,
                             concurrency_overhead=concurrency_overhead)

        result_queue = multiprocessing.Queue()
        event_queue = multiprocessing.Queue()

        def worker_args_gen(times_overhead, concurrency_overhead):
            """Generate arguments for process worker.

            Remainder of threads per process division is distributed to
            process workers equally - one thread per each process worker
            until the remainder equals zero. The same logic is applied
            to concurrency overhead.
            :param times_overhead: remaining number of threads to be
                                   distributed to workers
            :param concurrency_overhead: remaining number of maximum
                                         concurrent threads to be
                                         distributed to workers
            """
            while True:
                yield (
                    result_queue, iteration_gen, timeout,
                    times_per_worker + (times_overhead and 1),
                    concurrency_per_worker + (concurrency_overhead and 1),
                    context, cls, method_name, args, event_queue,
                    self.aborted, runs_per_second, self.config["rps"],
                    processes_to_start
                )
                if times_overhead:
                    times_overhead -= 1
                if concurrency_overhead:
                    concurrency_overhead -= 1

        process_pool = self._create_process_pool(
            processes_to_start, _worker_process,
            worker_args_gen(times_overhead, concurrency_overhead))
        self._join_processes(process_pool, result_queue, event_queue)



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Remove admin domain name

Revision ID: 32fada9b2fde
Revises: 5b983f0c9b9a
Create Date: 2016-08-29 08:32:30.818019

"""

from alembic import op
import sqlalchemy as sa

from rally.common.db import sa_types
from rally import exceptions

# revision identifiers, used by Alembic.
revision = "32fada9b2fde"
down_revision = "6ad4f426f005"
branch_labels = None
depends_on = None


deployments_helper = sa.Table(
    "deployments",
    sa.MetaData(),
    sa.Column("id", sa.Integer, primary_key=True, autoincrement=True),
    sa.Column(
        "config",
        sa_types.MutableJSONEncodedDict,
        default={},
        nullable=False,
    )
)


def upgrade():
    connection = op.get_bind()
    for deployment in connection.execute(deployments_helper.select()):
        conf = deployment.config
        if conf["type"] != "ExistingCloud":
            continue

        should_update = False

        if "admin_domain_name" in conf["admin"]:
            del conf["admin"]["admin_domain_name"]
            should_update = True
        if "users" in conf:
            for user in conf["users"]:
                if "admin_domain_name" in user:
                    del user["admin_domain_name"]
                    should_update = True

        if should_update:
            connection.execute(
                deployments_helper.update().where(
                    deployments_helper.c.id == deployment.id).values(
                    config=conf))


def downgrade():
    raise exceptions.DowngradeNotSupported()



# Copyright 2016: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import datetime as dt

import mock

from rally.common import objects
from rally import consts
from tests.unit import test


class VerificationTestCase(test.TestCase):
    def setUp(self):
        super(VerificationTestCase, self).setUp()

        self.db_obj = {"uuid": "uuid-1",
                       "env_uuid": "e_uuid"}
        self._db_entry = {}

    @mock.patch("rally.common.objects.verification.db.verification_create")
    def test_init(self, mock_verification_create):
        v = objects.Verification(self.db_obj)
        self.assertEqual(0, mock_verification_create.call_count)
        self.assertEqual(self.db_obj["uuid"], v.uuid)
        self.assertEqual(self.db_obj["uuid"], v["uuid"])

    def test_to_dict(self):
        TIME_FORMAT = "%Y-%m-%dT%H:%M:%S%z"
        data = {"created_at": dt.date(2017, 2, 3),
                "updated_at": dt.date(2017, 3, 3),
                "id": "v_id",
                "env_uuid": "d_uuid",
                "uuid": "v_uuid",
                "verifier_uuid": "v_uuid",
                "unexpected_success": "2",
                "status": "False",
                "tests": {"test1": "tdata1",
                          "test2": "tdata2"},
                "skipped": 2,
                "tests_duration": "",
                "tags": None,
                "run_args": "args",
                "success": 0,
                "expected_failures": 2,
                "tests_count": 3,
                "failures": 2}
        verification = objects.Verification(copy.deepcopy(data))
        result = verification.to_dict()
        data["created_at"] = data["created_at"].strftime(TIME_FORMAT)
        data["updated_at"] = data["updated_at"].strftime(TIME_FORMAT)
        data["deployment_uuid"] = data["env_uuid"]
        self.assertEqual(data, result)

    @mock.patch("rally.common.objects.verification.db.verification_create")
    def test_create(self, mock_verification_create):
        objects.Verification.create("some-verifier", "some-deployment", [], {})
        mock_verification_create.assert_called_once_with(
            "some-verifier", "some-deployment", [], {})

    @mock.patch("rally.common.objects.verification.db.verification_get")
    def test_get(self, mock_verification_get):
        mock_verification_get.return_value = self.db_obj
        v = objects.Verification.get(self.db_obj["uuid"])
        mock_verification_get.assert_called_once_with(self.db_obj["uuid"])
        self.assertEqual(self.db_obj["uuid"], v.uuid)

    @mock.patch("rally.common.objects.verification.db.verification_list")
    def test_list(self, mock_verification_list):
        mock_verification_list.return_value = [self.db_obj]
        vs = objects.Verification.list()
        mock_verification_list.assert_called_once_with(None, None, None, None)
        self.assertEqual(self.db_obj["uuid"], vs[0].uuid)

    @mock.patch("rally.common.objects.verification.db.verification_delete")
    def test_delete(self, mock_verification_delete):
        objects.Verification(self.db_obj).delete()
        mock_verification_delete.assert_called_once_with(self.db_obj["uuid"])

    @mock.patch("rally.common.objects.verification.db.verification_update")
    def test_update_status(self, mock_verification_update):
        v = objects.Verification(self.db_obj)
        v.update_status(status="some-status")
        mock_verification_update.assert_called_once_with(self.db_obj["uuid"],
                                                         status="some-status")

    @mock.patch("rally.common.objects.verification.db.verification_update")
    def test_finish(self, mock_verification_update):
        v = objects.Verification(self.db_obj)
        totals = {
            "tests_count": 2,
            "tests_duration": 0.54,
            "success": 2,
            "skip": 0,
            "expected_failures": 0,
            "unexpected_success": 0,
            "failures": 0
        }
        tests = {
            "foo_test[gate,negative]": {
                "name": "foo_test",
                "duration": 0.25,
                "status": "success",
                "tags": ["gate", "negative"]
            },
            "bar_test[gate,negative]": {
                "name": "bar_test",
                "duration": 0.29,
                "status": "success",
                "tags": ["gate", "negative"]
            }
        }
        v.finish(totals, tests)
        mock_verification_update.assert_called_once_with(
            self.db_obj["uuid"], status=consts.VerificationStatus.FINISHED,
            tests=tests, **totals)

        v = objects.Verification(self.db_obj)
        totals.update(failures=1)
        mock_verification_update.reset_mock()
        v.finish(totals, tests)
        mock_verification_update.assert_called_once_with(
            self.db_obj["uuid"], status=consts.VerificationStatus.FAILED,
            tests=tests, **totals)

        v = objects.Verification(self.db_obj)
        totals.update(failures=0, unexpected_success=1)
        mock_verification_update.reset_mock()
        v.finish(totals, tests)
        mock_verification_update.assert_called_once_with(
            self.db_obj["uuid"], status=consts.VerificationStatus.FAILED,
            tests=tests, **totals)

    @mock.patch("rally.common.objects.verification.db.verification_update")
    def test_set_error(self, mock_verification_update):
        v = objects.Verification(self.db_obj)
        v.set_error("Some error")
        mock_verification_update.assert_called_once_with(
            self.db_obj["uuid"], status=consts.VerificationStatus.CRASHED)



# Copyright 2015: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections

import mock

from rally.task import atomic
from tests.unit import test


class ActionTimerMixinTestCase(test.TestCase):

    def test_atomic_actions(self):
        inst = atomic.ActionTimerMixin()
        self.assertEqual(inst._atomic_actions, inst.atomic_actions())

    def test_reset_atomic_actions(self):
        inst = atomic.ActionTimerMixin()

        with atomic.ActionTimer(inst, "test"):
            pass
        self.assertNotEqual([], inst.atomic_actions())

        inst.reset_atomic_actions()
        self.assertEqual([], inst.atomic_actions())


class AtomicActionTestCase(test.TestCase):

    @mock.patch("time.time", side_effect=[1, 3, 6, 10, 15, 21])
    def test_action_timer_context(self, mock_time):
        inst = atomic.ActionTimerMixin()

        with atomic.ActionTimer(inst, "test"):
            with atomic.ActionTimer(inst, "test"):
                with atomic.ActionTimer(inst, "some"):
                    pass

        expected = [{"name": "test",
                     "started_at": 1,
                     "finished_at": 21,
                     "children": [{"name": "test",
                                   "started_at": 3,
                                   "finished_at": 15,
                                   "children": [{"name": "some",
                                                 "started_at": 6,
                                                 "finished_at": 10,
                                                 "children": []}]}]}]
        self.assertEqual(expected,
                         inst.atomic_actions())

    @mock.patch("time.time", side_effect=[1, 3])
    def test_action_timer_context_with_exception(self, mock_time):
        inst = atomic.ActionTimerMixin()

        class TestException(Exception):
            pass

        try:
            with atomic.ActionTimer(inst, "test"):
                raise TestException("test")
        except TestException:
            pass

        self.assertEqual([{"name": "test", "children": [], "failed": True,
                           "started_at": 1, "finished_at": 3}],
                         inst.atomic_actions())

    @mock.patch("time.time", side_effect=[1, 3])
    def test_action_timer_decorator(self, mock_time):

        class Some(atomic.ActionTimerMixin):

            @atomic.action_timer("some")
            def some_func(self, a, b):
                return a + b

        inst = Some()
        self.assertEqual(5, inst.some_func(2, 3))
        self.assertEqual([{"name": "some", "children": [],
                           "started_at": 1, "finished_at": 3}],
                         inst.atomic_actions())

    @mock.patch("time.time", side_effect=[1, 3])
    def test_action_timer_decorator_with_exception(self, mock_time):

        class TestException(Exception):
            pass

        class TestTimer(atomic.ActionTimerMixin):

            @atomic.action_timer("test")
            def some_func(self):
                raise TestException("test")

        inst = TestTimer()
        self.assertRaises(TestException, inst.some_func)
        self.assertEqual([{"name": "test", "children": [], "failed": True,
                           "started_at": 1, "finished_at": 3}],
                         inst.atomic_actions())

    def test_merge_atomic_actions(self):
        expected = [("foo", {"duration": 2, "count": 1,
                             "children": collections.OrderedDict()}),
                    ("bar", {"duration": 5, "count": 2,
                             "children": collections.OrderedDict()}),
                    ("do_something_bad", {
                        "duration": 1, "count": 1, "failed": True,
                        "children": collections.OrderedDict()})]
        result = atomic.merge_atomic_actions(
            [{"name": "foo", "started_at": 4, "finished_at": 6,
              "children": []},
             {"name": "bar", "started_at": 6, "finished_at": 8,
              "children": []},
             {"name": "bar", "started_at": 8, "finished_at": 11,
              "children": []},
             {"name": "do_something_bad", "started_at": 11, "finished_at": 12,
              "children": [], "failed": True}
             ])
        result = list(result.items())
        self.assertEqual(expected, result)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import subprocess

import mock

from rally import exceptions
from rally.plugins.common.verification import testr
from tests.unit import test


PATH = "rally.plugins.common.verification.testr"


class TestrContextTestCase(test.TestCase):

    def setUp(self):
        super(TestrContextTestCase, self).setUp()
        self.verifier = mock.Mock()
        self.prepare_run_args = self.verifier.manager.prepare_run_args
        self.prepare_run_args.side_effect = lambda x: x

    def assertEqualCmd(self, expected, actual, msg="", stestr=False):
        cmd = ["stestr" if stestr else "testr", "run", "--subunit"]
        cmd.extend(expected)
        self.assertEqual(cmd, actual, message=msg)

    def test_setup_with_concurrency(self):
        # default behaviour
        cfg = {"verifier": self.verifier}
        ctx = testr.TestrContext(cfg)
        ctx.setup()
        self.assertEqualCmd(["--parallel"], cfg["testr_cmd"])
        cfg = {"verifier": self.verifier, "run_args": {"concurrency": 0}}
        ctx = testr.TestrContext(cfg)
        ctx.setup()
        self.assertEqualCmd(["--parallel"], cfg["testr_cmd"])

        # serial mode
        cfg = {"verifier": self.verifier,
               "run_args": {"concurrency": 1}}
        ctx = testr.TestrContext(cfg)
        ctx.setup()
        self.assertEqualCmd(["--concurrency", "1"], cfg["testr_cmd"])

        # parallel mode
        cfg = {"verifier": self.verifier,
               "run_args": {"concurrency": 2}}
        ctx = testr.TestrContext(cfg)
        ctx.setup()
        self.assertEqualCmd(["--parallel", "--concurrency", "2"],
                            cfg["testr_cmd"])

    def test_setup_with_concurrency_stestr(self):
        self.verifier.manager._use_testr = False
        # default behaviour
        cfg = {"verifier": self.verifier}
        ctx = testr.TestrContext(cfg)
        ctx.setup()
        self.assertEqualCmd([], cfg["testr_cmd"], stestr=True)
        cfg = {"verifier": self.verifier, "run_args": {"concurrency": 0}}
        ctx = testr.TestrContext(cfg)
        ctx.setup()
        self.assertEqualCmd([], cfg["testr_cmd"], stestr=True)

        # serial mode
        cfg = {"verifier": self.verifier,
               "run_args": {"concurrency": 1}}
        ctx = testr.TestrContext(cfg)
        ctx.setup()
        self.assertEqualCmd(["--serial"], cfg["testr_cmd"], stestr=True)

        # parallel mode
        cfg = {"verifier": self.verifier,
               "run_args": {"concurrency": 2}}
        ctx = testr.TestrContext(cfg)
        ctx.setup()
        self.assertEqualCmd(["--concurrency", "2"], cfg["testr_cmd"],
                            stestr=True)

    @mock.patch("%s.common_utils.generate_random_path" % PATH)
    def test_setup_with_skip_and_load_lists(self, mock_generate_random_path):
        # with load_list, but without skip_list
        load_list = ["tests.foo", "tests.bar"]
        cfg = {"verifier": self.verifier,
               "run_args": {"load_list": load_list}}
        ctx = testr.TestrContext(cfg)
        mock_open = mock.mock_open()
        with mock.patch("%s.open" % PATH, mock_open):
            ctx.setup()
        mock_open.assert_called_once_with(
            mock_generate_random_path.return_value, "w")
        handle = mock_open.return_value
        handle.write.assert_called_once_with("\n".join(load_list))
        self.assertEqualCmd(["--parallel", "--load-list",
                             mock_generate_random_path.return_value],
                            cfg["testr_cmd"])
        self.assertFalse(self.verifier.manager.list_tests.called)

        # with load_list and skip_list
        load_list = ["tests.foo", "tests.bar"]
        skip_list = {"tests.foo": "skip reason"}
        cfg = {"verifier": self.verifier,
               "run_args": {"load_list": load_list,
                            "skip_list": skip_list}}
        ctx = testr.TestrContext(cfg)
        mock_open = mock.mock_open()
        with mock.patch("%s.open" % PATH, mock_open):
            ctx.setup()
        mock_open.assert_called_once_with(
            mock_generate_random_path.return_value, "w")
        handle = mock_open.return_value
        handle.write.assert_called_once_with(load_list[1])
        self.assertEqualCmd(["--parallel", "--load-list",
                             mock_generate_random_path.return_value],
                            cfg["testr_cmd"])
        self.assertFalse(self.verifier.manager.list_tests.called)

        # with skip_list, but without load_list
        load_list = ["tests.foo", "tests.bar"]
        self.verifier.manager.list_tests.return_value = load_list
        skip_list = {"tests.foo": "skip reason"}
        cfg = {"verifier": self.verifier,
               "run_args": {"skip_list": skip_list}}
        ctx = testr.TestrContext(cfg)
        mock_open = mock.mock_open()
        with mock.patch("%s.open" % PATH, mock_open):
            ctx.setup()
        mock_open.assert_called_once_with(
            mock_generate_random_path.return_value, "w")
        handle = mock_open.return_value
        handle.write.assert_called_once_with(load_list[1])
        self.assertEqualCmd(["--parallel", "--load-list",
                             mock_generate_random_path.return_value],
                            cfg["testr_cmd"])
        self.verifier.manager.list_tests.assert_called_once_with()

    @mock.patch("%s.common_utils.generate_random_path" % PATH)
    def test_skip_list_with_regex_positive_match(self,
                                                 mock_generate_random_path):
        # using a regex in skip_list
        load_list = ["tests.foo.bar", "tests.bar"]
        self.verifier.manager.list_tests.return_value = load_list
        skip_list = {"^tests.foo": "skip reason"}
        cfg = {"verifier": self.verifier,
               "run_args": {"skip_list": skip_list}}
        ctx = testr.TestrContext(cfg)
        mock_open = mock.mock_open()
        with mock.patch("%s.open" % PATH, mock_open):
            ctx.setup()
        mock_open.assert_called_once_with(
            mock_generate_random_path.return_value, "w")
        handle = mock_open.return_value
        handle.write.assert_called_once_with(load_list[1])
        self.assertEqualCmd(["--parallel", "--load-list",
                             mock_generate_random_path.return_value],
                            cfg["testr_cmd"])
        self.verifier.manager.list_tests.assert_called_once_with()

    @mock.patch("%s.common_utils.generate_random_path" % PATH)
    def test_skip_list_with_invalid_regex(self,
                                          mock_generate_random_path):
        load_list = [
            "tests.foo[e3976dea-bed9-4b14-abaf-59372de9303]",
            "tests.bar"
        ]
        self.verifier.manager.list_tests.return_value = load_list
        skip_list = {
            "tests.foo[e3976dea-bed9-4b14-abaf-59372de9303]": "skip reason"
        }
        cfg = {"verifier": self.verifier,
               "run_args": {"skip_list": skip_list}}
        ctx = testr.TestrContext(cfg)
        mock_open = mock.mock_open()
        with mock.patch("%s.open" % PATH, mock_open):
            ctx.setup()
        mock_open.assert_called_once_with(
            mock_generate_random_path.return_value, "w")
        handle = mock_open.return_value
        handle.write.assert_called_once_with(load_list[1])
        self.assertEqualCmd(["--parallel", "--load-list",
                             mock_generate_random_path.return_value],
                            cfg["testr_cmd"])
        self.verifier.manager.list_tests.assert_called_once_with()

    def test_setup_with_failing(self):
        cfg = {"verifier": self.verifier, "run_args": {"failed": True}}
        ctx = testr.TestrContext(cfg)
        ctx.setup()
        self.assertEqualCmd(["--parallel", "--failing"], cfg["testr_cmd"])

    def test_setup_with_pattern(self):
        cfg = {"verifier": self.verifier, "run_args": {"pattern": "foo"}}
        ctx = testr.TestrContext(cfg)
        ctx.setup()
        self.assertEqualCmd(["--parallel", "foo"], cfg["testr_cmd"])

    @mock.patch("%s.os.remove" % PATH)
    @mock.patch("%s.os.path.exists" % PATH)
    def test_cleanup(self, mock_exists, mock_remove):
        files = {"/path/foo_1": True,
                 "/path/bar_1": False,
                 "/path/foo_2": False,
                 "/path/bar_2": True}

        def fake_exists(path):
            return files.get(path, False)

        mock_exists.side_effect = fake_exists

        ctx = testr.TestrContext({"verifier": self.verifier})
        ctx._tmp_files = files.keys()

        ctx.cleanup()

        self.assertEqual([mock.call(f) for f in files.keys()],
                         mock_exists.call_args_list)
        self.assertEqual([mock.call(f) for f in files.keys() if files[f]],
                         mock_remove.call_args_list)


class TestrLauncherTestCase(test.TestCase):
    def test_run_environ_property(self):
        env = mock.Mock()

        class FakeLauncher(testr.TestrLauncher):
            @property
            def environ(self):
                return env

        self.assertEqual(env, FakeLauncher(mock.Mock()).run_environ)

    @mock.patch("%s.utils.check_output" % PATH)
    def test_list_tests_via_stestr(self, mock_check_output):
        mock_check_output.return_value = (
            "logging message\n"  # should be ignored
            "one more useless data\n"  # should be ignored
            "tests.FooTestCase.test_something\n"  # valid
            "tests.FooTestCase.test_another[\n"  # invalid
            "tests.BarTestCase.test_another[id=123]\n"  # valid
            "tests.FooTestCase.test_another[id=a2-213,smoke]\n"  # valid
        )
        verifier = mock.Mock()

        launcher = testr.TestrLauncher(verifier)
        launcher._use_testr = False

        self.assertEqual(["tests.FooTestCase.test_something",
                          "tests.BarTestCase.test_another[id=123]",
                          "tests.FooTestCase.test_another[id=a2-213,smoke]"],
                         launcher.list_tests())

        mock_check_output.assert_called_once_with(
            ["stestr", "list", ""],
            cwd=launcher.repo_dir, env=launcher.environ, debug_output=False)

    @mock.patch("%s.utils.check_output" % PATH)
    def test_list_tests_via_testr(self, mock_check_output):
        mock_check_output.return_value = (
            "logging message\n"  # should be ignored
            "one more useless data\n"  # should be ignored
            "tests.FooTestCase.test_something\n"  # valid
            "tests.FooTestCase.test_another[\n"  # invalid
            "tests.BarTestCase.test_another[id=123]\n"  # valid
            "tests.FooTestCase.test_another[id=a2-213,smoke]\n"  # valid
        )
        verifier = mock.Mock()

        launcher = testr.TestrLauncher(verifier)
        launcher._use_testr = True

        self.assertEqual(["tests.FooTestCase.test_something",
                          "tests.BarTestCase.test_another[id=123]",
                          "tests.FooTestCase.test_another[id=a2-213,smoke]"],
                         launcher.list_tests())

        mock_check_output.assert_called_once_with(
            ["testr", "list-tests", ""],
            cwd=launcher.repo_dir, env=launcher.environ, debug_output=False)

    @mock.patch("%s.shutil.rmtree" % PATH)
    @mock.patch("%s.utils.check_output" % PATH)
    @mock.patch("%s.os.path.exists" % PATH)
    @mock.patch("%s.os.path.isdir" % PATH)
    def test__init_testr(self, mock_isdir, mock_exists, mock_check_output,
                         mock_rmtree):
        launcher = testr.TestrLauncher(mock.Mock())
        mock_exists.assert_called_once_with(
            os.path.join(launcher.repo_dir, ".testr.conf"))
        mock_exists.reset_mock()

        # case #1: testr already initialized
        mock_isdir.return_value = True

        launcher._init_testr()

        self.assertFalse(mock_check_output.called)
        self.assertFalse(mock_exists.called)
        self.assertFalse(mock_rmtree.called)

        # case #2: initializing testr without errors
        mock_isdir.return_value = False

        launcher._init_testr()

        mock_check_output.assert_called_once_with(
            ["testr", "init"], cwd=launcher.repo_dir, env=launcher.environ)
        self.assertFalse(mock_exists.called)
        self.assertFalse(mock_rmtree.called)
        mock_check_output.reset_mock()

        # case #3: initializing stestr without errors
        launcher._use_testr = False

        launcher._init_testr()

        mock_check_output.assert_called_once_with(
            ["stestr", "init"], cwd=launcher.repo_dir, env=launcher.environ)
        self.assertFalse(mock_exists.called)
        self.assertFalse(mock_rmtree.called)
        mock_check_output.reset_mock()

        # case #4: initializing testr with error
        mock_check_output.side_effect = OSError
        test_repository_dir = os.path.join(launcher.base_dir,
                                           ".testrepository")

        self.assertRaises(exceptions.RallyException, launcher._init_testr)

        mock_check_output.assert_called_once_with(
            ["stestr", "init"], cwd=launcher.repo_dir, env=launcher.environ)
        mock_exists.assert_called_once_with(test_repository_dir)
        mock_rmtree.assert_called_once_with(test_repository_dir)

    @mock.patch("%s.subunit_v2.parse" % PATH)
    @mock.patch("%s.subprocess.Popen" % PATH)
    def test_run(self, mock_popen, mock_parse):
        launcher = testr.TestrLauncher(mock.Mock())
        ctx = {"testr_cmd": ["ls", "-la"],
               "xfail_list": None, "skip_list": None}

        self.assertEqual(mock_parse.return_value, launcher.run(ctx))

        mock_popen.assert_called_once_with(ctx["testr_cmd"],
                                           env=launcher.run_environ,
                                           cwd=launcher.repo_dir,
                                           stdout=subprocess.PIPE,
                                           stderr=subprocess.STDOUT)
        mock_popen.return_value.wait.assert_called_once_with()
        mock_parse.assert_called_once_with(
            mock_popen.return_value.stdout, live=True,
            expected_failures=None,
            skipped_tests=None,
            logger_name=launcher.verifier.name)

    @mock.patch("%s.subunit_v2.parse" % PATH)
    @mock.patch("%s.subprocess.Popen" % PATH)
    def test_run_skip(self, mock_popen, mock_parse):
        launcher = testr.TestrLauncher(mock.Mock())
        skip_list = {"test1": "reason1"}
        ctx = {"testr_cmd": ["ls", "-la"],
               "xfail_list": None, "skip_list": skip_list}

        self.assertEqual(mock_parse.return_value, launcher.run(ctx))

        mock_popen.assert_called_once_with(ctx["testr_cmd"],
                                           env=launcher.run_environ,
                                           cwd=launcher.repo_dir,
                                           stdout=subprocess.PIPE,
                                           stderr=subprocess.STDOUT)
        mock_popen.return_value.wait.assert_called_once_with()
        mock_parse.assert_called_once_with(
            mock_popen.return_value.stdout, live=True,
            expected_failures=None,
            skipped_tests=ctx["skip_list"],
            logger_name=launcher.verifier.name)

    @mock.patch("%s.subunit_v2.parse" % PATH)
    @mock.patch("%s.subprocess.Popen" % PATH)
    def test_run_exclude(self, mock_popen, mock_parse):
        launcher = testr.TestrLauncher(mock.Mock())
        xfail_list = {"test1": "reason1"}
        ctx = {"testr_cmd": ["ls", "-la"],
               "xfail_list": xfail_list, "skip_list": None}
        launcher = testr.TestrLauncher(mock.Mock())

        result = launcher.run(ctx)
        self.assertEqual(mock_parse.return_value, result)

        mock_popen.assert_called_once_with(ctx["testr_cmd"],
                                           env=launcher.run_environ,
                                           cwd=launcher.repo_dir,
                                           stdout=subprocess.PIPE,
                                           stderr=subprocess.STDOUT)
        mock_popen.return_value.wait.assert_called_once_with()
        mock_parse.assert_called_once_with(
            mock_popen.return_value.stdout, live=True,
            expected_failures=xfail_list,
            skipped_tests=None,
            logger_name=launcher.verifier.name)

    @mock.patch("%s.manager.VerifierManager.install" % PATH)
    def test_install(self, mock_verifier_manager_install):
        launcher = testr.TestrLauncher(mock.Mock())
        launcher._init_testr = mock.Mock()

        launcher.install()

        mock_verifier_manager_install.assert_called_once_with()
        launcher._init_testr.assert_called_once_with()



# Copyright (c) 2016 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Refactor task results

Revision ID: e654a0648db0
Revises: 3177d36ea270
Create Date: 2016-04-01 14:36:56.373349

"""
import datetime as dt
import json
import uuid

from alembic import op
import sqlalchemy as sa

from rally.common.db import sa_types
from rally import exceptions

# revision identifiers, used by Alembic.
revision = "e654a0648db0"
down_revision = "32fada9b2fde"
branch_labels = None
depends_on = None


taskhelper = sa.Table(
    "tasks",
    sa.MetaData(),
    sa.Column("created_at", sa.DateTime(), nullable=True),
    sa.Column("updated_at", sa.DateTime(), nullable=True),
    sa.Column("id", sa.Integer(), nullable=False),
    sa.Column("uuid", sa.String(length=36), nullable=False),
    sa.Column("status", sa.Enum(
        "aborted", "aborting", "cleaning up", "failed", "finished",
        "init", "paused", "running", "setting up", "soft_aborting",
        "verifying", name="enum_tasks_status"), nullable=False),
    sa.Column("verification_log", sa.Text(), nullable=True),
    sa.Column("tag", sa.String(length=64), nullable=True),
    sa.Column("deployment_uuid", sa.String(length=36), nullable=False),
    sa.Column("title", sa.String(length=64), default=""),
    sa.Column("description", sa.Text(), default=""),
    sa.Column("input_task", sa.Text(), default=""),
    sa.Column("validation_duration", sa.Float()),
    sa.Column("task_duration", sa.Float()),
    sa.Column("pass_sla", sa.Boolean()),
    sa.Column(
        "validation_result",
        sa_types.MutableJSONEncodedDict(),
        default={},
        nullable=False
    )
)

task_result_helper = sa.Table(
    "task_results",
    sa.MetaData(),
    sa.Column("created_at", sa.DateTime()),
    sa.Column("updated_at", sa.DateTime()),
    sa.Column("id", sa.Integer(), nullable=False, autoincrement=True),
    sa.Column(
        "key",
        sa_types.MutableJSONEncodedDict(),
        nullable=False),
    sa.Column(
        "data",
        sa_types.MutableJSONEncodedDict(),
        nullable=False),
    sa.Column("task_uuid", sa.String(length=36), nullable=True)
)

taghelper = sa.Table(
    "tags",
    sa.MetaData(),
    sa.Column("created_at", sa.DateTime()),
    sa.Column("updated_at", sa.DateTime()),
    sa.Column("id", sa.Integer(), nullable=False, autoincrement=True),
    sa.Column("uuid", sa.String(length=36), nullable=False),
    sa.Column("tag", sa.String(length=255), nullable=False),

    sa.Column(
        "type",
        sa.Enum(
            "task", "subtask",
            name="enum_tag_types"),
        nullable=False)
)


def upgrade():
    conn = op.get_bind()

    subtask_table = op.create_table(
        "subtasks",
        sa.Column("created_at", sa.DateTime()),
        sa.Column("updated_at", sa.DateTime()),
        sa.Column("id", sa.Integer(), nullable=False, autoincrement=True),
        sa.Column("uuid", sa.String(length=36), nullable=False),
        sa.Column("task_uuid", sa.String(length=36), nullable=False),
        sa.Column("title", sa.String(length=64), default=""),
        sa.Column("description", sa.Text(), default=""),
        sa.Column(
            "context",
            sa_types.MutableJSONEncodedDict(),
            default={},
            nullable=False),

        sa.Column(
            "sla",
            sa_types.MutableJSONEncodedDict(),
            default={},
            nullable=False),

        sa.Column("duration", sa.Float()),

        sa.Column(
            "run_in_parallel",
            sa.Boolean(),
            default=False,
            nullable=False),

        sa.Column("pass_sla", sa.Boolean()),

        sa.Column(
            "status",
            sa.Enum(
                "finished", "running", "crashed",
                name="enum_subtasks_status"),
            nullable=False),

        sa.ForeignKeyConstraint(["task_uuid"], ["tasks.uuid"], ),
        sa.PrimaryKeyConstraint("id")

    )

    op.create_index("subtask_uuid", "subtasks", ["uuid"], unique=True)
    op.create_index("subtask_status", "subtasks", ["status"], unique=False)

    workload_table = op.create_table(
        "workloads",
        sa.Column("created_at", sa.DateTime()),
        sa.Column("updated_at", sa.DateTime()),
        sa.Column("id", sa.Integer(), nullable=False, autoincrement=True),
        sa.Column("uuid", sa.String(length=36), nullable=False),
        sa.Column("task_uuid", sa.String(length=36), nullable=False),
        sa.Column("subtask_uuid", sa.String(length=36), nullable=False),
        sa.Column("name", sa.String(length=64), nullable=False),
        sa.Column("description", sa.Text(), default=""),
        sa.Column("position", sa.Integer(), default=0, nullable=False),

        sa.Column(
            "runner_type",
            sa.String(length=64),
            nullable=False),

        sa.Column(
            "runner",
            sa_types.MutableJSONEncodedDict(),
            default={},
            nullable=False),

        sa.Column(
            "args",
            sa_types.MutableJSONEncodedDict(),
            default={},
            nullable=False),

        sa.Column(
            "context",
            sa_types.MutableJSONEncodedDict(),
            default={},
            nullable=False),

        sa.Column(
            "hooks",
            sa_types.MutableJSONEncodedList(),
            default=[],
            nullable=False),

        sa.Column(
            "sla",
            sa_types.MutableJSONEncodedDict(),
            default={},
            nullable=False),

        sa.Column(
            "sla_results",
            sa_types.MutableJSONEncodedDict(),
            default={},
            nullable=False),

        sa.Column(
            "context_execution",
            sa_types.MutableJSONEncodedDict(),
            default={},
            nullable=False),

        sa.Column("load_duration", sa.Float(), default=0),
        sa.Column("full_duration", sa.Float(), default=0),
        sa.Column("min_duration", sa.Float(), default=0),
        sa.Column("max_duration", sa.Float(), default=0),
        sa.Column("total_iteration_count", sa.Integer(), default=0),
        sa.Column("failed_iteration_count", sa.Integer(), default=0),

        sa.Column("pass_sla", sa.Boolean()),

        sa.Column(
            "statistics",
            sa_types.MutableJSONEncodedDict(),
            default={},
            nullable=False),


        sa.Column("start_time", sa.DateTime()),
        sa.Column("_profiling_data", sa.Text(), default=""),

        sa.ForeignKeyConstraint(["task_uuid"], ["tasks.uuid"], ),
        sa.ForeignKeyConstraint(["subtask_uuid"], ["subtasks.uuid"], ),
        sa.PrimaryKeyConstraint("id")
    )

    op.create_index("workload_uuid", "workloads", ["uuid"], unique=True)

    workloaddata_table = op.create_table(
        "workloaddata",
        sa.Column("created_at", sa.DateTime()),
        sa.Column("updated_at", sa.DateTime()),
        sa.Column("id", sa.Integer(), nullable=False, autoincrement=True),
        sa.Column("uuid", sa.String(length=36), nullable=False),
        sa.Column("task_uuid", sa.String(length=36), nullable=False),
        sa.Column("workload_uuid", sa.String(length=36), nullable=False),
        sa.Column("chunk_order", sa.Integer(), nullable=False),
        sa.Column("iteration_count", sa.Integer(), nullable=False),
        sa.Column("failed_iteration_count", sa.Integer(), nullable=False),
        sa.Column("chunk_size", sa.Integer(), nullable=False),

        sa.Column(
            "compressed_chunk_size",
            sa.Integer(),
            nullable=False),

        sa.Column("started_at", sa.DateTime(), nullable=False),
        sa.Column("finished_at", sa.DateTime(), nullable=False),
        sa.Column(
            "chunk_data",
            sa_types.MutableJSONEncodedDict(),
            default={},
            nullable=False),

        sa.ForeignKeyConstraint(["task_uuid"], ["tasks.uuid"], ),
        sa.ForeignKeyConstraint(["workload_uuid"], ["workloads.uuid"], ),
        sa.PrimaryKeyConstraint("id")
    )

    op.create_index(
        "workload_data_uuid", "workloaddata", ["uuid"], unique=True)

    tag_table = op.create_table(
        "tags",
        sa.Column("created_at", sa.DateTime()),
        sa.Column("updated_at", sa.DateTime()),
        sa.Column("id", sa.Integer(), nullable=False, autoincrement=True),
        sa.Column("uuid", sa.String(length=36), nullable=False),
        sa.Column("tag", sa.String(length=255), nullable=False),

        sa.Column(
            "type",
            sa.Enum(
                "task", "subtask",
                name="enum_tag_types"),
            nullable=False),

        sa.PrimaryKeyConstraint("id")
    )

    op.create_index(
        "d_type_tag", "tags", ["uuid", "type", "tag"], unique=True)

    with op.batch_alter_table("tasks") as batch_op:
        batch_op.add_column(
            sa.Column("title", sa.String(length=64), default="")
        )

        batch_op.add_column(
            sa.Column("description", sa.Text(), default="")
        )

        batch_op.add_column(
            sa.Column("input_task", sa.Text(), default="")
        )

        batch_op.add_column(
            sa.Column("validation_duration", sa.Float())
        )

        batch_op.add_column(
            sa.Column("task_duration", sa.Float())
        )

        batch_op.add_column(
            sa.Column("pass_sla", sa.Boolean())
        )

        batch_op.add_column(
            sa.Column(
                "validation_result",
                sa_types.MutableJSONEncodedDict(),
                default={})
        )

    for task in conn.execute(taskhelper.select()):
        if task.tag:
            conn.execute(
                tag_table.insert(),
                [{
                    "uuid": task.uuid,
                    "type": "task",
                    "tag": task.tag,
                    "created_at": task.created_at,
                    "updated_at": task.updated_at
                }]
            )

        task_results = conn.execute(
            task_result_helper.select().
            where(task_result_helper.c.task_uuid == task.uuid)
        )

        pass_sla = True
        task_duration = 0

        for task_result in task_results:
            raw_data = task_result.data.get("raw", [])
            iter_count = len(raw_data)

            failed_iter_count = 0
            max_duration = 0
            min_duration = -1

            for d in raw_data:
                if d.get("error"):
                    failed_iter_count += 1

                duration = d.get("duration", 0)

                if duration > max_duration:
                    max_duration = duration

                if min_duration < 0 or min_duration > duration:
                    min_duration = duration

            sla = task_result.data.get("sla", [])
            success = all([s.get("success") for s in sla])

            if not success:
                pass_sla = False

            task_duration += task_result.data.get("full_duration", 0)

            delta = dt.timedelta(
                seconds=task_result.data.get("full_duration", 0))
            start = task_result.created_at - delta

            subtask_uuid = str(uuid.uuid4())

            conn.execute(
                subtask_table.insert(),
                [{
                    "uuid": subtask_uuid,
                    "task_uuid": task.uuid,
                    "created_at": task_result.created_at,
                    "updated_at": task_result.updated_at,
                    # NOTE(ikhudoshyn) We don't have info on subtask status
                    "status": "finished",
                    "duration": task_result.data.get("full_duration", 0),
                    "pass_sla": success
                }]
            )

            workload_uuid = str(uuid.uuid4())

            conn.execute(
                workload_table.insert(),
                [{
                    "created_at": task_result.created_at,
                    "updated_at": task_result.updated_at,
                    "uuid": workload_uuid,
                    "task_uuid": task.uuid,
                    "subtask_uuid": subtask_uuid,
                    "name": task_result.key["name"],
                    "position": task_result.key["pos"],
                    "runner_type": task_result.key["kw"]["runner"]["type"],
                    "runner": task_result.key["kw"]["runner"],
                    "context": task_result.key["kw"].get("context", {}),
                    "sla": task_result.key["kw"].get("sla", {}),
                    "args": task_result.key["kw"].get("args", {}),
                    "sla_results": {"sla": sla},
                    "context_execution": {},
                    "load_duration": task_result.data.get("load_duration", 0),
                    "full_duration": task_result.data.get("full_duration", 0),
                    "min_duration": min_duration,
                    "max_duration": max_duration,
                    "total_iteration_count": iter_count,
                    "failed_iteration_count": failed_iter_count,
                    "pass_sla": success,
                    "statistics": {},
                    "start_time": start,
                }]
            )

            conn.execute(
                workloaddata_table.insert(),
                [{
                    "uuid": str(uuid.uuid4()),
                    "task_uuid": task.uuid,
                    "workload_uuid": workload_uuid,
                    "chunk_order": 0,
                    "iteration_count": iter_count,
                    "failed_iteration_count": failed_iter_count,
                    "chunk_data": {"raw": raw_data},
                    # TODO(ikhudoshyn)
                    "chunk_size": 0,
                    "compressed_chunk_size": 0,
                    "started_at": start,
                    "finished_at": task_result.created_at
                }]
            )

        task_verification_log = {}
        if task.verification_log:
            task_verification_log = json.loads(task.verification_log)

        conn.execute(
            taskhelper.update().where(taskhelper.c.uuid == task.uuid),
            {
                "pass_sla": pass_sla,
                "task_duration": task_duration,
                "validation_duration": 0,
                "validation_result": task_verification_log
            }
        )

    # TODO(ikhudoshyn) update workload's statistics

    with op.batch_alter_table("tasks") as batch_op:
        batch_op.drop_column("tag")
        batch_op.drop_column("verification_log")
        batch_op.alter_column(
            "validation_result",
            existing_type=sa_types.MutableJSONEncodedDict(),
            nullable=False)

    op.drop_table("task_results")


def downgrade():
    raise exceptions.DowngradeNotSupported()



# Copyright 2014: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import ddt

from rally.plugins.common.sla import max_average_duration
from rally.task import sla
from tests.unit import test


@ddt.ddt
class MaxAverageDurationTestCase(test.TestCase):

    @ddt.data((1, True), (1000, True), (0, False))
    @ddt.unpack
    def test_validate(self, config, valid):
        results = sla.SLA.validate("max_avg_duration", None, None, config)
        if valid:
            self.assertEqual([], results)
        else:
            self.assertEqual(1, len(results))

    def test_result(self):
        sla1 = max_average_duration.MaxAverageDuration(42)
        sla2 = max_average_duration.MaxAverageDuration(3.62)
        for sla_inst in [sla1, sla2]:
            sla_inst.add_iteration({"duration": 3.14})
            sla_inst.add_iteration({"duration": 6.28})
        self.assertTrue(sla1.result()["success"])   # 42 > avg([3.14, 6.28])
        self.assertFalse(sla2.result()["success"])  # 3.62 < avg([3.14, 6.28])
        self.assertEqual("Passed", sla1.status())
        self.assertEqual("Failed", sla2.status())

    def test_result_no_iterations(self):
        sla_inst = max_average_duration.MaxAverageDuration(42)
        self.assertTrue(sla_inst.result()["success"])

    def test_add_iteration(self):
        sla_inst = max_average_duration.MaxAverageDuration(4.0)
        self.assertTrue(sla_inst.add_iteration({"duration": 3.5}))
        self.assertTrue(sla_inst.add_iteration({"duration": 2.5}))
        self.assertTrue(
            sla_inst.add_iteration({"duration": 5.0}))   # avg = 3.667
        self.assertFalse(
            sla_inst.add_iteration({"duration": 7.0}))  # avg = 4.5
        self.assertTrue(
            sla_inst.add_iteration({"duration": 1.0}))   # avg = 3.8

    @ddt.data([[1.0, 2.0, 1.5, 4.3],
               [2.1, 3.4, 1.2, 6.3, 7.2, 7.0, 1.],
               [1.1, 1.1, 2.2, 2.2, 3.3, 4.3]])
    def test_merge(self, durations):

        single_sla = max_average_duration.MaxAverageDuration(4.0)

        for dd in durations:
            for d in dd:
                single_sla.add_iteration({"duration": d})

        slas = [max_average_duration.MaxAverageDuration(4.0)
                for _ in durations]

        for idx, sla_inst in enumerate(slas):
            for duration in durations[idx]:
                sla_inst.add_iteration({"duration": duration})

        merged_sla = slas[0]
        for sla_inst in slas[1:]:
            merged_sla.merge(sla_inst)

        self.assertEqual(single_sla.success, merged_sla.success)
        self.assertEqual(single_sla.avg, merged_sla.avg)



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Change tag type enum

Revision ID: a6f364988fc2
Revises: 37fdbb373e8d
Create Date: 2017-01-17 18:47:10.700459

"""
from alembic import op
import sqlalchemy as sa

from rally import exceptions

# revision identifiers, used by Alembic.
revision = "a6f364988fc2"
down_revision = "37fdbb373e8d"
branch_labels = None
depends_on = None


TAG_TYPES = ["task", "subtask"]

tag_helper = sa.Table(
    "tags",
    sa.MetaData(),
    sa.Column("id", sa.Integer(), nullable=False),
    sa.Column("type", sa.Enum(*TAG_TYPES, name="enum_tag_types"),
              nullable=False),
    sa.Column("new_type", sa.String(36), nullable=False)
)


def upgrade():
    with op.batch_alter_table("tags") as batch_op:
        batch_op.add_column(
            sa.Column("new_type", sa.String(36)))

    op.execute(tag_helper.update().values(new_type=tag_helper.c.type))

    op.drop_index("d_type_tag", "tags")

    with op.batch_alter_table("tags") as batch_op:
        batch_op.drop_column("type")
        batch_op.alter_column("new_type", new_column_name="type",
                              existing_type=sa.String(36), nullable=False)

    op.create_index("d_type_tag", "tags", ["uuid", "type", "tag"], unique=True)


def downgrade():
    raise exceptions.DowngradeNotSupported()



# Copyright (C) 2014 Yahoo! Inc. All Rights Reserved.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from rally.common.plugin import plugin
from rally.task import scenario
from rally.task import types
from tests.unit import test


class ConvertTestCase(test.TestCase):

    # NOTE(stpierre): These cases test types.convert(),
    # types._get_preprocessor_loader(), and bits of
    # types.preprocess(). This may not look very elegant, but it's the
    # easiest way to test both convert() and
    # _get_preprocessor_loader() without getting so fine-grained that
    # the tests are basically tests that the computer is on.

    def setUp(self):
        super(ConvertTestCase, self).setUp()

        @types.convert(bar={"type": "test_bar"})
        @scenario.configure(name="FakeConvertPlugin.one_arg")
        class FakeConvertOneArgPlugin(scenario.Scenario):
            def run(self, bar):
                """Dummy docstring.

                :param bar: dummy parameter
                """
                pass

        self.addCleanup(FakeConvertOneArgPlugin.unregister)

        @types.convert(bar={"type": "test_bar"},
                       baz={"type": "test_baz"})
        @scenario.configure(name="FakeConvertPlugin.two_args")
        class FakeConvertTwoArgsPlugin(scenario.Scenario):

            def run(self, bar, baz):
                """Dummy docstring.

                :param bar: dummy parameter
                :param baz: dummy parameter
                """
                pass

        self.addCleanup(FakeConvertTwoArgsPlugin.unregister)

    @mock.patch("rally.task.types.ResourceType.get", create=True)
    def test_convert(self, mock_resource_type_get):
        ctx = mock.MagicMock()
        args = types.preprocess("FakeConvertPlugin.one_arg",
                                ctx,
                                {"bar": "bar_config"})
        mock_resource_type_get.assert_called_once_with("test_bar")
        resourcetype_cls = mock_resource_type_get.return_value
        resourcetype_cls.assert_called_once_with(ctx, {})
        resourcetype_obj = resourcetype_cls.return_value
        resourcetype_obj.pre_process.assert_called_once_with(
            resource_spec="bar_config", config={"type": "test_bar"})
        self.assertDictEqual(
            args, {"bar": resourcetype_obj.pre_process.return_value})

    @mock.patch("rally.task.types.ResourceType.get", create=True)
    def test_convert_multiple(self, mock_resource_type_get):
        resourcetype_classes = {"bar": mock.Mock(), "baz": mock.Mock()}

        def _get_resource_type(name):
            # cut "test_" prefix
            name = name[5:]
            if name in resourcetype_classes:
                return resourcetype_classes[name]
            self.fail("The unexpected resource class tried to be used.")

        mock_resource_type_get.side_effect = _get_resource_type

        ctx = mock.MagicMock()
        scenario_args = {"bar": "bar_config", "baz": "baz_config"}
        processed_args = types.preprocess(
            "FakeConvertPlugin.two_args", ctx, scenario_args)

        mock_resource_type_get.assert_has_calls([mock.call("test_bar"),
                                                 mock.call("test_baz")],
                                                any_order=True)

        expected_dict = {}
        for resourcetype_n, resourcetype_cls in resourcetype_classes.items():
            resourcetype_cls.assert_called_once_with(ctx, {})
            resourcetype_obj = resourcetype_cls.return_value
            resourcetype_obj.pre_process.assert_called_once_with(
                resource_spec=scenario_args[resourcetype_n],
                config={"type": "test_%s" % resourcetype_n})
            return_value = resourcetype_obj.pre_process.return_value
            expected_dict[resourcetype_n] = return_value

        self.assertDictEqual(expected_dict, processed_args)


class PreprocessTestCase(test.TestCase):

    @mock.patch("rally.task.types.scenario.Scenario.get")
    def test_preprocess(self, mock_scenario_get):

        name = "some_plugin"
        type_name = "%s_type" % self.id()

        context = {
            "a": 1,
            "b": 2,
        }
        args = {"a": 10, "b": 20}

        @plugin.configure(type_name)
        class SomeType(types.ResourceType):

            def pre_process(self, resource_spec, config):
                return resource_spec * 2

        mock_scenario_get.return_value._meta_get.return_value = {
            "a": {"type": type_name}
        }

        result = types.preprocess(name, context, args)
        mock_scenario_get.assert_called_once_with(name)
        mock_scenario_get.return_value._meta_get.assert_called_once_with(
            "preprocessors", default={})
        self.assertEqual({"a": 20, "b": 20}, result)
